"""Base class for Integration objects
"""

from rockset.swagger_client.api import IntegrationsApi

class Integration(object):
    TYPE_AWS = 'AWS'

    @classmethod
    def list(cls, **kwargs):
        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage. '
                'use rockset.Client().Integration.list() instead.'
            )
        client = kwargs.pop('client')

        integrations = IntegrationsApi(client).list().get('data')

        list = []
        for i in integrations:
            type = cls.TYPE_AWS
            list.append(cls(client=client, type=type, **i.to_dict()))
        return list

    @classmethod
    def retrieve(cls, **kwargs):
        """Retrieves a single integration

        Args:
            name (str): Name of the integration to be retrieved

        Returns:
            Integration: Integration object
        """

        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage. '
                'use rockset.Client().Integration.retrieve() instead.'
            )

        client = kwargs.pop('client')
        name = kwargs.pop('name')
        integration = IntegrationsApi(client).get(integration=name).get('data')
        return cls(client=client, type=cls.TYPE_AWS, **integration.to_dict())

    # instance_methods
    def __init__(self, client, name, type, **kwargs):
        """Represents a single Integration"""
        self.client = client
        self.name = name
        self.type = type
        for key in kwargs:
            setattr(self, key, kwargs[key])
        return

    def __str__(self):
        """Converts the collection into a user friendly printable string"""
        return str(vars(self))

    def asdict(self):
        d = vars(self)
        d.pop('client')
        return d

    def drop(self):
        IntegrationsApi(self.client).delete(integration=self.name)
        return



"""
Introduction
------------
Integration objects represents a single Rockset integration.
These objects are generally created using a Rockset Client
object using methods such as::

    from rockset import Client

    # connect to Rockset
    rs = Client(api_key=...)

    # create a new integration
    aws_integration = rs.Integration.AWS.create('aws-integration')
"""
from .integration import Integration

from rockset.swagger_client.api import IntegrationsApi
from rockset.swagger_client.models import (AwsKeyIntegration, CreateIntegrationRequest)

class AWSIntegration(Integration):
    @classmethod
    def create(cls, name, aws_access_key_id, aws_secret_access_key, **kwargs):

        if aws_access_key_id is None or aws_secret_access_key is None:
            raise ValueError(
                'aws_access_key_id and aws_secret_access_key required '
                'to create an AWS integration'
            )

        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage.'
                'use rockset.Client().Integrations.AWS.create() instead'
            )

        client = kwargs.pop('client')

        aws_credentials = AwsKeyIntegration(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key)

        request = CreateIntegrationRequest(
                name=name,
                aws=aws_credentials,
                **kwargs
                )

        integration = IntegrationsApi(client).create(body=request).get('data')
        return cls(client, **integration.to_dict())

    def __init__(self, *args, **kwargs):
        kwargs['type'] = Integration.TYPE_AWS
        super().__init__(*args, **kwargs)



"""
Introduction
------------
Cursor objects are return by the Collection.query_ API.
A cursor simply binds the query to a particular collection and the
query is not executed server side until the application starts to fetch
results from the cursor.

.. note:: Cursors are never instantiated directly by the application \
and are always instantiated by Collection ``query()`` APIs.

Fetch all results
-----------------

Use the cursor's results() method to fetch all the results of the query
in one shot::

    results = mycollection.query(q).results()

The above technique will work well, if the number of results returned by the
query is limited, say because it employs a LIMIT clause.

For queries, that return a large number of results, please use the cursor
iterators as described below.

Iterators with automatic pagination
-----------------------------------

Cursor objects are iterables, so you can do something like::

    results = mycollection.query(q)
    for r in results:
        print(r)

Cursor objects support seamless automatic pagination to iterate over large
result sets. The cursor iterator will fetch and buffer a small portion of the
results and as the iterator reaches the end of the current batch of buffered
results, it will automatically issue the query with the appropriate pagination
parameters to buffer the next batch and seamlessly continue results iteration.

The default cursor iterator uses a batch size of 10,000. You can create a cursor
iterator with a different batch size by using the ``iter()`` method.

Example using the default cursor iterator::

    results = mycollection.query(q)
    for r in results:
        print(r)


Example using a custom cursor iterator with batch size 200::

    results = mycollection.query(q)
    for r in results.iter(200):
        print(r)

.. automethod:: rockset.Cursor.iter()

Async requests
--------------

Cursors support asyncio.Future to schedule and run queries concurrently
along with other async events in your application.

One can create an asyncio.Future from a cursor object using the
``Cursor.async_request()`` method. These futures are not scheduled in
any async threads and the application have to schedule them in an asyncio
event loop. Once the futures are scheduled and run to completion, then
the results of their respective queries can be accessed by calling
future.result(). The return value of future.result() will be identical
to calling Cursor.results() API on the original query.

For example::

    jims_future = people.query(F["first_name"] == "Jim").async_request()
    sfmerch_future = merchants.query(F["zipcode"] == "94107").async_request()

    # can schedule these asyncio.futures along with other futures
    # issue both queries concurrently and block until both of them finish
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(jims_future, sfmerch_future))

    all_jims = jims_future.result()
    all_sfmerchants = sfmerch_future.result()

Even if an future was originally issued by an async_requst() API call,
one can still call the blocking Cursor.results() API to fetch the results
synchronously. Cursor.results() will schedule the future, block until the
future finishes execution and then will return those results. A subsequent
future.result() call will return the query results immediately.

For example::

    jims_cursor = people.query(F["first_name"] == "Jim")
    jims_future = jims_cursor.async_request()

    # do a blocking results() will block on the future behind the scenes
    results = jims_cursor.results()

    # this will return immeidately without incurring any server round-trip
    results2 = jims_future.result()


.. automethod:: rockset.Cursor.async_request()

"""

from rockset.exception import ServerError
from rockset.query import LimitQuery, SubQuery
from rockset.swagger_client.api.queries_api import QueriesApi
from rockset.swagger_client.models.query_request import QueryRequest
from rockset.swagger_client.models.query_response import QueryResponse
import asyncio


class Cursor(object):
    """ Fetch the results of a query executed against a collection
    """

    def __init__(self, q=None, client=None):
        self.query = q
        self.client = client
        self.all_results = None
        self.future = None

    def _run(self, limit=None, skip=0):
        if self.all_results is not None:
            start = skip
            if limit is None:
                stop = None
            else:
                stop = skip + limit
            return self.all_results[start:stop]
        return self._run_query(limit, skip)

    def _run_query(self, limit=None, skip=0):
        # add limit, offset to query if required
        query = self.query
        if limit is not None:
            query = LimitQuery(limit=limit, skip=skip, child=SubQuery(query))

        # build the query request object
        qo = {}
        (sqltxt, params) = query.sql()
        qo = {'query': sqltxt, 'parameters': params.sqlparams()}

        # execute the query
        req = QueryRequest(sql=qo)
        results = QueriesApi(self.client).query(req)
        if not isinstance(results, QueryResponse):
            raise ServerError(
                message='invalid return message (type={} results={})'.
                format(type(results), results)
            )
        return results.get('results')

    def __iter__(self):
        return self.iter()

    def __str__(self):
        return str(self.results())

    def __getitem__(self, key):
        return self.results()[key]

    def iter(self, batch=10000):
        """ Returns an iterator that does seamless automatic pagination
        behind the scenes while fetching no more than the specified
        batch size number of results at a time.

        Args:
            batch (int): maximum number of results fetched at a time

        Returns:
            Iterator Object: Iterator that will return all the results with \
seamless automatic pagination
        """
        if batch <= 0:
            raise ValueError(
                'invalid cursor iterator batchsize. "{}" is not '
                'a positive integer.'.format(batch)
            )

        skip = 0
        while True:
            results = self._run(batch, skip)
            for r in results:
                yield r
            if ((batch is None) or (batch > len(results))):
                break
            skip += batch

    def results(self):
        """ Execute the query and fetch all the results in one shot.

        Returns:
            Array of dicts: All the query result documents
        """
        if self.all_results is None:
            self.all_results = self._run()

        return self.all_results

    def async_request(self):
        """ Returns an asyncio.Future object that can be scheduled in an
        asyncio event loop. Once scheduled and run to completion, the results
        can be fetched via the future.result() API. The return value of
        future.result() will be the same as the return value of Cursor.results()

        Returns:
            asyncio.Future: Returns a Future object that can be scheduled
            in an asyncio event loop and future.result() will hold the same
            return value as Cursor.results()
        """
        if self.future is not None:
            return self.future
        self.future = asyncio.Future()
        asyncio.ensure_future(self._issue())
        return self.future

    async def _issue(self):
        self.all_results = self._run()
        self.future.set_result(self.all_results)


__all__ = [
    'Cursor',
]



# coding: utf-8

# flake8: noqa

"""
    REST API

    Rockset's REST API allows for creating and managing all resources in Rockset. Each supported endpoint is documented below.  All requests must be authorized with a Rockset API key, which can be created in the [Rockset console](https://console.rockset.com). The API key must be provided as `ApiKey <api_key>` in the `Authorization` request header. For example: ``` Authorization: ApiKey aB35kDjg93J5nsf4GjwMeErAVd832F7ad4vhsW1S02kfZiab42sTsfW5Sxt25asT ```  All endpoints are only accessible via https.  Build something awesome!  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import apis into sdk package
from rockset.swagger_client.api.api_keys_api import ApiKeysApi
from rockset.swagger_client.api.collections_api import CollectionsApi
from rockset.swagger_client.api.documents_api import DocumentsApi
from rockset.swagger_client.api.integrations_api import IntegrationsApi
from rockset.swagger_client.api.organizations_api import OrganizationsApi
from rockset.swagger_client.api.queries_api import QueriesApi
from rockset.swagger_client.api.users_api import UsersApi

# import ApiClient
from rockset.swagger_client.api_client import ApiClient
from rockset.swagger_client.configuration import Configuration
# import models into sdk package
from rockset.swagger_client.models.add_documents_request import AddDocumentsRequest
from rockset.swagger_client.models.add_documents_response import AddDocumentsResponse
from rockset.swagger_client.models.api_key import ApiKey
from rockset.swagger_client.models.aws_external_id_integration import AwsExternalIdIntegration
from rockset.swagger_client.models.aws_key_integration import AwsKeyIntegration
from rockset.swagger_client.models.commit_mark import CommitMark
from rockset.swagger_client.models.commit_mark_positions import CommitMarkPositions
from rockset.swagger_client.models.create_api_key_request import CreateApiKeyRequest
from rockset.swagger_client.models.create_api_key_response import CreateApiKeyResponse
from rockset.swagger_client.models.create_collection_request import CreateCollectionRequest
from rockset.swagger_client.models.create_collection_response import CreateCollectionResponse
from rockset.swagger_client.models.create_integration_request import CreateIntegrationRequest
from rockset.swagger_client.models.create_integration_response import CreateIntegrationResponse
from rockset.swagger_client.models.create_user_request import CreateUserRequest
from rockset.swagger_client.models.create_user_response import CreateUserResponse
from rockset.swagger_client.models.csv_params import CsvParams
from rockset.swagger_client.models.delete_api_key_response import DeleteApiKeyResponse
from rockset.swagger_client.models.delete_collection_response import DeleteCollectionResponse
from rockset.swagger_client.models.delete_documents_request import DeleteDocumentsRequest
from rockset.swagger_client.models.delete_documents_request_data import DeleteDocumentsRequestData
from rockset.swagger_client.models.delete_documents_response import DeleteDocumentsResponse
from rockset.swagger_client.models.delete_integration_response import DeleteIntegrationResponse
from rockset.swagger_client.models.delete_user_response import DeleteUserResponse
from rockset.swagger_client.models.document_status import DocumentStatus
from rockset.swagger_client.models.error_model import ErrorModel
from rockset.swagger_client.models.error_model_context import ErrorModelContext
from rockset.swagger_client.models.event_time_info import EventTimeInfo
from rockset.swagger_client.models.field_mapping import FieldMapping
from rockset.swagger_client.models.field_mapping_v2 import FieldMappingV2
from rockset.swagger_client.models.field_mask import FieldMask
from rockset.swagger_client.models.field_mask_mask import FieldMaskMask
from rockset.swagger_client.models.format_params import FormatParams
from rockset.swagger_client.models.get_collection_response import GetCollectionResponse
from rockset.swagger_client.models.get_integration_response import GetIntegrationResponse
from rockset.swagger_client.models.input_field import InputField
from rockset.swagger_client.models.integration import Integration
from rockset.swagger_client.models.list_api_keys_response import ListApiKeysResponse
from rockset.swagger_client.models.list_collections_response import ListCollectionsResponse
from rockset.swagger_client.models.list_integrations_response import ListIntegrationsResponse
from rockset.swagger_client.models.list_users_response import ListUsersResponse
from rockset.swagger_client.models.organization import Organization
from rockset.swagger_client.models.organization_response import OrganizationResponse
from rockset.swagger_client.models.output_field import OutputField
from rockset.swagger_client.models.query_field_type import QueryFieldType
from rockset.swagger_client.models.query_parameter import QueryParameter
from rockset.swagger_client.models.query_request import QueryRequest
from rockset.swagger_client.models.query_request_sql import QueryRequestSql
from rockset.swagger_client.models.query_response import QueryResponse
from rockset.swagger_client.models.query_response_stats import QueryResponseStats
from rockset.swagger_client.models.resource import Resource
from rockset.swagger_client.models.resource_stats import ResourceStats
from rockset.swagger_client.models.source import Source
from rockset.swagger_client.models.source_dynamo_db import SourceDynamoDb
from rockset.swagger_client.models.source_kinesis import SourceKinesis
from rockset.swagger_client.models.source_s3 import SourceS3
from rockset.swagger_client.models.sql_expression import SqlExpression
from rockset.swagger_client.models.user import User
from rockset.swagger_client.models.xml_params import XmlParams




import yaml
from .command_auth import AuthCommand
from datetime import timedelta
from docopt import docopt


class Create(AuthCommand):
    def usage(self, subcommand=None):
        usage = """
usage:
  rock create --help
  rock create collection --help
  rock create integration --help
  rock create [-h] --file=YAMLFILE
  rock create collection <name> [-h] [options] [<data_source_url> ...]
  rock create integration <name> --type=INTEGRATION_TYPE [-h] [options]


commands:
  collection          create a new collection.
                      you can optionally specify data sources to automatically
                      feed into the collection such as AWS S3.

  integration         create a new integration.
                      an integration object can store access details and
                      credentials of an external account (eg: an AWS account)
                      and can be used at collection creation time to access
                      one or more data sources.
                      integration objects allows you to securely store access
                      credentials and share across the team without actually
                      exposing the access credentials and secrets.

options for `rock create`:
  -d TEXT, --desc=TEXT        human readable description of the new resource
  -f FILE, --file=FILE        create all resources specified in the YAML file,
                              run `rock -o yaml describe <collection>` on an
                              existing collection to see the YAML format
  -h, --help                  show this help message and exit
        """
        usage_subcommand = {
            "collection":
                """
arguments for `rock create collection`:
  <name>              name of the new collection you wish to create
  <data_source_url>   specify the data source to auto ingest in order to
                      populate the collection
                      eg: s3://my-precious-s3-bucket
                          s3://my-precious-s3-bucket/data/path/prefix
                          dynamodb://my-precious-dynamodb-table-name
                          kinesis://my-precious-kinesis-stream-name

options for `rock create collection`:
  --integration=INTEGRATION_NAME                Specify an integration that will be used to access
                                                the source of this collection. For sources that don't need
                                                special access or credentials, this can be left unspecified.
  --event-time-field=FIELD_NAME                 specify a root-level field, representing time in one of the
                                                formats supported by rockset (look at --event-time-format)
                                                this field will be mapped to event-time.
  --event-time-format=TIME_FORMAT               specify the format of time in which the field mapped to event-time
                                                is formatted in your documents. (requires --event-time-field)
                                                supported formats   * milliseconds_since_epoch (default)
                                                                    * seconds_since_epoch
  --event-time-default-timezone=TIME_ZONE       specify the time zone of event times in the documents which will be
                                                added into this collection. (requires --event-time-field)
                                                default is 'UTC' (supported time zones: one of standard IANA time zones)
  --retention=RETENTION_DURATION                specify the minimum time duration using short-hand notation
                                                (such as 24h, 8d, or 13w) for which documents in this collection
                                                will be retained before being automatically deleted.
                                                (default: none i.e., documents will be retained indefinitely)
  --format=CSV                                  Specify the format of data in this source. If the data is
                                                comma separted values per line, then specify CSV. Other
                                                formats of data are auto detected and this parameter
                                                should not be specified for non-csv data formats.
  --csv-separator=","                           Separator for columns for csv files.
                                                This is applicable only when --format = CSV
  --csv-encoding="UTF-8"                        Encoding, one of "UTF-8", "UTF-16", "ISO-8859-1".
                                                This is applicable  only when --format = CSV.
  --csv-first-line-as-column-names=false        If true, then use the first column as the column names.
                                                This is applicable  only when --format = CSV.
  --csv-column-names="c1,col2",                 A comma separated list of column names
                                                This is applicable  only when --format = CSV.
  --csv-column-types="int,boolean",             A comma separated list of column types and should have a
                                                one-to-one mapping to the names specified in column-names.
                                                This is applicable  only when --format = CSV.
  --csv-schema-file="/home/schema.yaml"         A file that specifies the schema of the data.
                                                It is an yaml file that specifies the name of each column
                                                and the type of data in each column. If a schema file is
                                                specified, then the values specified by csv-column-names
                                                and csv-column-types are ignored.


examples:

    Create a collection and source all contents from an AWS S3 bucket:

        $ rock create collection customers s3://customers-mycompany-com

    Create a collection from an AWS S3 bucket but only pull a particular
    path prefix within the S3 bucket:

        $ rock create collection event-log \\
            s3://event-log.mycompany.com/root/path/in/bkt --integration aws-rockset-readonly

    Create a collection with a source that ingests data from kinesis stream:

        $ rock create collection events kinesis://click-streams --integration aws-rockset-readonly

    Create a collection and map a field to event-time

        $ rock create collection \\
            my-event-data --event-time-field timestamp --event-time-format milliseconds_since_epoch

    Create a collection with retention set to 10 days

        $ rock create collection \\
            my-event-data --retention="10d"

        """,
            "integration":
                """
arguments for `rock create integration`:
  <name>              name of the new integration you wish to create

options for `rock create integration --type=AWS`:
  --aws_access_key_id=AWS_ACCESS_KEY_ID              AWS access key id
  --aws_secret_access_key=AWS_SECRET_ACCESS_KEY      AWS secret access key

examples:

    Create an integration of type AWS

        $ rock create integration aws-rockset-readonly --type=AWS --aws_access_key_id=access_key --aws_secret_access_key=secret_access
        """
        }

        if subcommand == 'collection':
            return usage + usage_subcommand['collection']
        elif subcommand == 'integration':
            return usage + usage_subcommand['integration']
        elif subcommand == 'all':
            return (
                usage + usage_subcommand['collection'] +
                usage_subcommand['integration']
            )

        return usage

    def convert_to_seconds(self, duration):

        num = duration[:-1]
        try:
            num = int(num)
        except ValueError as e:
            ret = 'invalid duration "{}"\n'.format(num)
            raise ValueError(ret)

        unit = duration[-1]
        try:
            if unit == 'h':
                time_delta = timedelta(hours=num)
            elif unit == 'd':
                time_delta = timedelta(days=num)
            elif unit == 'w':
                time_delta = timedelta(weeks=num)
            else:
                ret = 'invalid time unit "{}"\n'.format(unit)
                raise ValueError(ret)
        except OverflowError:
            ret = 'duration "{}" too large for specified time units\n'.format(
                num
            )
            raise OverflowError(ret)

        return int(time_delta.total_seconds())

    def _source_s3(self, s3_url, integration, format_params):
        parts = s3_url[5:].split('/')
        bucket = parts[0]
        prefixes = (len(parts) > 1) and ['/'.join(parts[1:])] or []

        return self.client.Source.s3(
            bucket=bucket,
            prefixes=prefixes,
            integration=integration,
            format_params=format_params
        )

    def _source_dynamo(self, url, integration):
        table_name = url[11:]

        return self.client.Source.dynamo(
            table_name=table_name,
            integration=integration
        )

    def _source_kinesis(self, kinesis_url, integration):
        stream_name = kinesis_url[10:]

        return self.client.Source.kinesis(
            stream_name=stream_name,
            integration=integration
        )

    def parse_args(self, args):
        parsed_args = dict(docopt(self.usage('all'), argv=args, help=False))

        # handle help
        if parsed_args['--help']:
            if parsed_args['collection']:
                ret = self.usage('collection')
            elif parsed_args['integration']:
                ret = self.usage('integration')
            else:
                ret = self.usage()
            raise SystemExit(ret.strip())

        # see if YAMLFILE was specified
        fn = parsed_args['--file']
        if fn:
            self.set_batch_items('resource', self._parse_yaml_file(fn))
            return {}

        # construct a valid CreateRequest object
        resource = {}

        if parsed_args['collection']:
            resource['name'] = parsed_args['<name>']
            sources = []
            if parsed_args['--desc']:
                resource['description'] = parsed_args['--desc']

            integration = parsed_args['--integration']
            if integration is not None:
                integration = self.client.Integration.retrieve(integration)

            # Specifiy the format of data in collection source
            data_format = "auto_detect"
            format_params = None
            if parsed_args['--format'] is not None:
                data_format = parsed_args['--format']

            if data_format == "CSV":
                csv_separator = None
                csv_encoding = None
                csv_first_line_as_column_names = None
                csv_column_names = None
                csv_column_types = None
                if parsed_args['--csv-separator'] is not None:
                    csv_separator = parsed_args['--csv-separator']
                if parsed_args['--csv-encoding'] is not None:
                    csv_encoding = parsed_args['--csv-encoding']
                if parsed_args['--csv-first-line-as-column-names'] is not None:
                    csv_first_line_as_column_names = parsed_args[
                        '--csv-first-line-as-column-names']
                if parsed_args['--csv-column-names'] is not None:
                    csv_column_names_str = parsed_args['--csv-column-names']
                    csv_column_names = csv_column_names_str.split(',')
                if parsed_args['--csv-column-types'] is not None:
                    csv_column_types_str = parsed_args['--csv-column-types']
                    csv_column_types = csv_column_types_str.split(',')

                # if this is a csv schema file, load it and extract
                # the column names and types from it.
                if parsed_args['--csv-schema-file'] is not None:
                    csv_schema_file = parsed_args['--csv-schema-file']
                    with open(csv_schema_file, 'r') as stream:
                        data_loaded = yaml.load(stream)

                        # fetch appropriate entry for this collection from the schema file
                        all_columns = data_loaded.get('fields')
                        if all_columns is None:
                            print(
                                'No fields entry ' + resource['name'] +
                                ' in file ' + csv_schema_file
                            )
                            raise SystemExit

                        # one entry in the file shows up as
                        # [{'col1': 'integer'}, {'col2': 'string'}, {'col3': 'boolean'}]

                        csv_column_names = []
                        csv_column_types = []
                        for i in range(len(all_columns)):
                            onecol = all_columns[i]
                            assert len(onecol) == 1
                            colnames = list(all_columns[i].keys())
                            csv_column_names.append(list(onecol.keys())[0])
                            csv_column_types.append(list(onecol.values())[0])

                format_params = self.client.Source.csv_params(
                    separator=csv_separator,
                    encoding=csv_encoding,
                    first_line_as_column_names=csv_first_line_as_column_names,
                    column_names=csv_column_names,
                    column_types=csv_column_types
                )

            resource['type'] = 'COLLECTION'
            for source in parsed_args['<data_source_url>']:
                if source[:5] == 's3://':
                    sources.append(
                        self._source_s3(
                            source, integration, format_params
                        )
                    )
                elif source[:11] == 'dynamodb://':
                    sources.append(
                        self._source_dynamo(
                            source, integration
                        )
                    )
                elif source[:10] == 'kinesis://':
                    sources.append(
                        self._source_kinesis(
                            source, integration
                        )
                    )
                else:
                    ret = 'Error: invalid data source URL "{}"\n'.format(source)
                    ret += self.usage()
                    ret += self.usage('collection')
                    raise SystemExit(ret.strip())

            # handle options related to event_time
            if parsed_args['--event-time-field'] is not None:
                resource['event_time_field'] = parsed_args['--event-time-field']
                resource['event_time_format'] = parsed_args[
                    '--event-time-format']
                resource['event_time_default_timezone'] = parsed_args[
                    '--event-time-default-timezone']
            elif parsed_args['--event-time-format'] is not None:
                ret = 'Error: --event-time-field is required to specify --event-time-format'
                ret += self.usage()
                raise SystemExit(ret.strip())
            elif parsed_args['--event-time-default-timezone'] is not None:
                ret = 'Error: --event-time-field is required to specify --event-time-default-timezone'
                ret += self.usage()
                raise SystemExit(ret.strip())

            #handle retention
            if parsed_args['--retention'] is not None:
                try:
                    retention = self.convert_to_seconds(
                        parsed_args['--retention']
                    )
                    resource['retention_secs'] = retention
                except ValueError as e:
                    ret = 'Error: invalid argument "{}" for --retention, {}'.format(
                        parsed_args['--retention'], str(e)
                    )
                    ret += self.usage()
                    raise SystemExit(ret)
                except OverflowError as e:
                    ret = 'Error: invalid value "{}" for --retention, {}'.format(
                        parsed_args['--retention'], str(e)
                    )
                    ret += self.usage()
                    raise SystemExit(ret)

            resource['sources'] = sources
            return {'resource': resource}
        elif parsed_args['integration']:
            resource['name'] = parsed_args['<name>']
            resource['type'] = 'INTEGRATION'

            if parsed_args['--desc']:
                resource['description'] = parsed_args['--desc']

            if parsed_args['--type'] == 'AWS':
                resource['integration_type'] = 'AWS'

                resource['aws_access_key_id'] = parsed_args[
                    '--aws_access_key_id']
                resource['aws_secret_access_key'] = parsed_args[
                    '--aws_secret_access_key']
            else:
                ret = "Error: invalid integration type, supported types: AWS\n"
                ret += self.usage()
                ret += self.usage('integration')
                raise SystemExit(ret.strip())

        return {'resource': resource}

    def go(self):
        self.logger.info('create {}'.format(self.resource))
        rtype = self.resource.pop('type', None)
        if rtype is None:
            return 1
        if rtype == 'COLLECTION':
            return self.go_collection(self.resource)
        elif rtype == 'INTEGRATION':
            return self.go_integration(self.resource)
        return 1

    def go_collection(self, resource):
        name = resource.pop('name')
        c = self.client.Collection.create(name, **resource)
        self.lprint(0, 'Collection "%s" was created successfully.' % (c.name))
        return 0

    def go_integration(self, resource):
        name = resource.pop('name')
        integration_type = resource.pop('integration_type')

        if integration_type == 'AWS':
            aws_access_key_id = resource.pop('aws_access_key_id')
            aws_secret_access_key = resource.pop('aws_secret_access_key')
            try:
                i = self.client.Integration.AWS.create(
                    name, aws_access_key_id, aws_secret_access_key, **resource
                )
                self.lprint(
                    0, 'Integration "%s" was created successfully.' % (i.name)
                )
            except ValueError as e:
                ret = "Error: {}\n".format(str(e))
                ret += self.usage('integration')
                raise SystemExit(ret.strip())
            return 0
        return 1



from .dbapi import connect, Connection, Cursor, DictCursor
from .sqlalchemy import RocksetDialect

__all__ = [
  # DBAPI
  connect,
  Connection,
  Cursor,
  DictCursor,

  # SQL Alchemy
  RocksetDialect,
]



from .command_auth import AuthCommand
from .sql.rscli.main import cli_main
from rockset.query import QueryStringSQLText


class SQLQuery(AuthCommand):
    def usage(self):
        return """
usage: rock sql --help
       rock sql
       rock sql <sql_statement> [<args>...]

Run a sql query and return results as documents.

arguments:
  <sql-statement>       sql query to run; will read from STDIN if == '-'

examples:

  # To enter into the rock SQL REPL
  $ rock sql

  # To run a simple SQL query
  $ rock sql 'SELECT * from my_collection LIMIT 10'

  # To supply SQL from STDIN, use "-"
  $ echo 'SELECT * from my_collection LIMIT 10' | rock sql -

        """

    def validate_args(self, pargs):
        allowed_args = ['flood']
        for arg in pargs['<args>']:
            if arg not in allowed_args:
                return False
        return True

    def go(self):
        if not self.sql_statement:
            cli_main(
                api_server=self.api_server,
                api_key=self.api_key,
                workspace='commons'
            )
            return 0
        elif self.sql_statement == '-':
            self.sql_statement = self.read_stdin('SQL query')
        q = QueryStringSQLText(self.sql_statement)

        # handle advanced args
        self.flood = 'flood' in self.args

        # lets do this
        results = self.client.sql(q=q, flood_all_leaves=self.flood).results()

        self.print_list(0, results)
        return 0



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rockset/document_locator.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rockset/document_locator.proto',
  package='',
  syntax='proto3',
  serialized_options=_b('\n\020io.rockset.protoP\000'),
  serialized_pb=_b('\n\x1erockset/document_locator.proto\"7\n\x14\x44ocumentLocatorProto\x12\x0e\n\x06raw_id\x18\x01 \x01(\x04\x12\x0f\n\x07indices\x18\x02 \x03(\rB\x14\n\x10io.rockset.protoP\x00\x62\x06proto3')
)




_DOCUMENTLOCATORPROTO = _descriptor.Descriptor(
  name='DocumentLocatorProto',
  full_name='DocumentLocatorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_id', full_name='DocumentLocatorProto.raw_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='indices', full_name='DocumentLocatorProto.indices', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=89,
)

DESCRIPTOR.message_types_by_name['DocumentLocatorProto'] = _DOCUMENTLOCATORPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DocumentLocatorProto = _reflection.GeneratedProtocolMessageType('DocumentLocatorProto', (_message.Message,), dict(
  DESCRIPTOR = _DOCUMENTLOCATORPROTO,
  __module__ = 'rockset.document_locator_pb2'
  # @@protoc_insertion_point(class_scope:DocumentLocatorProto)
  ))
_sym_db.RegisterMessage(DocumentLocatorProto)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



"""
Introduction
------------
FieldMapping objects represent various transformations that apply to all
documents inserted into a collection.

Example usage
-------------
::
    from rockset import Client

    rs = Client()

    mappings = [
        rs.FieldMapping.mapping(
            name="anonymize_name"
            input_fields=[
                    {
                        rs.FieldMapping.input_fields(
                            field_name="name", if_missing="SKIP",
                            is_drop=true, param="name")
                    }
                ]
            output_field=rs.FieldMapping.output_field(
                field_name="name_anon",
                sql_expression="SHA256(:name)",
                on_error="FAIL")
            )]
    newcoll = rs.Collection.create(name="newcoll", field_mappings=mappings)
"""


class FieldMapping(object):
    def __str__(self):
        return str(vars(self))

    def __iter__(self):
        for k, v in vars(self).items():
            yield (k, v)

    @classmethod
    def mapping(cls, name, input_fields=None, output_field=None):
        """Creates a new mapping object

        Args:
            name(str): Name of the mapping
            input_fields: An Array of InputField objects
            output: An OutputField object
        """
        input_dict = [dict(f) for f in input_fields]
        output_dict = dict(output_field)
        return Mapping(
            name=name, input_fields=input_dict, output_field=output_dict
        )

    @classmethod
    def input_field(cls, field_name, param=None, if_missing=None, is_drop=None):
        """Create a new InputField object

        Args:
            field_name (str): The name of the field, parsed as a SQL qualified name
            param (str): SQL parameter name (default: same as field name. Required
                if the field name is nested)
            if_missing (str): Define the behavior if field name is missing or is NULL,
                one of:
                    "SKIP" (skip the update),
                    "PASS" (pass NULL to the update function)
                [default: "SKIP"]
            is_drop (boolean): Set to true if the input field needs to be dropped
        """
        return InputField(
            field_name=field_name,
            param=param,
            if_missing=if_missing,
            is_drop=is_drop
        )

    @classmethod
    def output_field(cls, field_name, sql_expression, on_error=None):
        """Create a new OutputField object

        Args:
            name (str): The name of the field, parsed as SQL qualified name
            value (Value): SQL expression
            on_error (str): Define the error behavior, one of:
                "SKIP" (skip output field),
                "FAIL" (fail the update)
                [default: "SKIP"]
            """
        return OutputField(
            field_name=field_name,
            sql_expression=sql_expression,
            on_error=on_error
        )


class Mapping(FieldMapping):
    def __init__(self, name, input_fields=None, output_field=None):
        self.name = name
        self.input_fields = input_fields

        if output_field is not None:
            self.output_field = output_field


class InputField:
    def __init__(self, field_name, param=None, if_missing=None, is_drop=None):
        self.field_name = field_name
        if param is not None:
            self.param = param
        if if_missing is not None:
            self.if_missing = if_missing
        if is_drop is not None:
            self.is_drop = is_drop
        return

    def __str__(self):
        return str(vars(self))

    def __iter__(self):
        for k, v in vars(self).items():
            yield (k, v)


class OutputField:
    def __init__(self, field_name, sql_expression, on_error=None):
        self.field_name = field_name
        self.value = {
            'sql': sql_expression,
        }
        if on_error is not None:
            self.on_error = on_error
        return

    def __str__(self):
        return str(vars(self))

    def __iter__(self):
        for k, v in vars(self).items():
            yield (k, v)



# -*- coding: utf-8 -*-
import sys

from docopt import docopt
from rockset import version
from rockset.credentials import Credentials
from rockset.exception import AuthError

def main(args=None):
    # init args from sys
    if args is None:
        args = sys.argv[1:]

    usage = """usage: rock [--version] [--help]
       rock [-v | -vv | -vvv] [options] <command> [<args> ...]
       rock help <command>

Command line interface to Rockset

Options:
  --version                  display rockset version and exit

  -h, --help                 print help message and exit
  -o, --format=FORMAT        select output format; one of {json,text,yaml}
                             [default: text]
  -p, --profile=PROFILE      select name of the credentials profile to use
  -v, --verbose              increase output verbosity

Commands:

  Setup Credentials:
       configure             setup auth credentials via API keys

  Data Definition:
       create                create a new collection or integration
       describe              get detailed stats about a collection or integration
       drop                  drop an existing collection or integration
       ls                    list all collections or integrations

  Data Management:
       get                   retrieve documents from a collection
       rm                    remove a document from a collection
       sql                   run a SQL query from command-line or
                             enter SQL REPL when run without any args
       upload                upload local files to a collection

  Other Commands:
       help                  more information on a specific command
       play                  play a game of {rock,paper,scissors}


Use 'rock help <command>' for more information on a specific command.
"""
    parsed_args = docopt(usage, argv=args, help=False, version=version(),
        options_first=True)

    # setup command and args
    command = parsed_args['<command>']
    argv = [command] + parsed_args['<args>']

    # setup credentials
    Credentials().setup()

    # process the command - delayed imports save load time for every run
    if command == 'configure':
        from rockset.rock.commands.configure import Configure
        command_cls = Configure
    elif command == 'create':
        from rockset.rock.commands.create import Create
        command_cls = Create
    elif command == 'describe':
        from rockset.rock.commands.describe import Describe
        command_cls = Describe
    elif command == 'drop':
        from rockset.rock.commands.drop import Drop
        command_cls = Drop
    elif command == 'ls':
        from rockset.rock.commands.list import List
        command_cls = List
    elif command == 'upload':
        from rockset.rock.commands.upload import Upload
        command_cls = Upload
    elif command == 'get':
        from rockset.rock.commands.get import Get
        command_cls = Get
    elif command == 'rm':
        from rockset.rock.commands.remove import Remove
        command_cls = Remove
    elif command == 'sql':
        from rockset.rock.commands.sqlquery import SQLQuery
        command_cls = SQLQuery
    elif command == 'play':
        from rockset.rock.commands.play import Play
        command_cls = Play
    elif command == 'help':
        argv = parsed_args['<args>'] + ['--help']
        main(args=argv)
        return 0
    else:
        print(usage)
        return 0

    # catch exceptions in constructor
    try:
        c = command_cls(**parsed_args)
    except AuthError as e:
        print('Error: {}'.format(str(e)), file=sys.stderr)
        print('Hint: Use "rock configure" to update your credentials.',
            file=sys.stderr)
        return 1
    except Exception as e:
        print('Error: {} {}'.format(type(e).__name__, str(e)), file=sys.stderr)
        return 1

    # let Command.main_go() deal with all exceptions here
    return c.main(argv)


if __name__ == '__main__':
    main()



"""
Usage
-----

Query module contains a set of APIs that allows you to compose powerful
queries over collections.

This module comprises of two major components:

* ``Q`` : Query Builder
    Used to compose complex and powerful queries.

>>> from rockset import Q
>>> q = Q('hello-world').limit(10)
>>> (sqltxt, sqlargs) = q.sql()
>>> print(sqltxt)
SELECT *
FROM "hello-world"
LIMIT 10
>>>

* ``F`` : Field Reference
    Used to construct field expressions that refer to particular fields
    within a document.

>>> from rockset import F
>>> (F['answer'] == 42).sqlexpression()
'"answer" = 42'
>>>

Example
-------
::

    from rockset import Client, Q, F

    # connect to Rockset
    api_key = 'adkjf234rksjfa23waejf2'
    rs = Client(api_key=...)

    # fetch user whose "_id" == "u42"
    u = rs..query(Q('users').where(F["_id"] == "u42"))

    # fetch the 100 oldest users in the 'users' collection
    q = Q('users').highest(100, F["age"])
    old100 = rs.query(q)

    # find the average rating of all songs by "The Beatles"
    q = Q('songs').where(F["artist"] == "The Beatles").select(F["rating"].avg())
    avg_rating = rs.query(q)

Query Operators: Overview
-------------------------

.. autofunction:: rockset.Q

* Constructor

  * ``Q``: Specify the collection to be queried

  >>> # return all documents in the logins collection.
  >>> q = Q('logins')

* Filter queries

  * ``where``: Classic selection operator to only return documents that match \
the given criteria. Use ``F`` to construct field expressions to specify the \
selection criteria.

  >>> # return all docs in logins where field "user_id" is equal to "u42"
  >>> q = Q('logins').where(F['user_id'] == 'u42')

* Projection

  * ``select``: Specify the list of desired fields to be returned

  >>> # will return the fields "user_id" and "login_ip" from all docs in logins
  >>> q = Q('logins').select(F['user_id'], F['login_ip'])

* Pagination

  * ``limit`` : Specify limit with skip support

  >>> # return 10 documents from logins after skipping the first 40 results
  >>> q = Q('logins').limit(10, skip=40)

* Sorting

  * ``highest``, ``lowest``: Find the top N or the bottom N

  >>> # will return 10 documents with the most recent "login_time"
  >>> q = Q('logins').highest(10, F['login_time'])

* Aggregation

  * ``aggregate``: Group by and aggregate fields

  >>> # will aggregate all documents in logins by "user_id",
  >>> # and return "user_id", max("login_time") and count(*) after aggregation.
  >>> Q('logins').aggregate(F['user_id'], F['login_time'].max(), F.count())

* Joins

  * ``join``:  Regular JOIN operator
  * TODO: docs coming soon
  * ``lookup``: LEFT OUTER JOIN operator
  * TODO: docs coming soon
  * ``apply``: Graph traversal operator
  * TODO: docs coming soon

Field Expressions Overview
--------------------------

.. autodata:: rockset.F

* Value comparators

  * ``==``, ``!=``, ``<``, ``<=``, ``>``, ``>=``:

  >>> # match all docs where "first_name" is equal to "Jim"
  >>> F["first_name"] == "Jim"
  >>> # match all docs where "rating" is greater than or equal to 4.5
  >>> F["rating"] >= 4.5
  >>> # match all docs where "title" text is lexographcially greater than "Star Wars"
  >>> F["title"] >= "Star Wars"

* String functions

  * ``startswith``, ``like``: Prefix and classic SQL LIKE expressions

  >>> # match all docs where "title" starts with "Star Wars"
  >>> F["title"].startswith("Stars Wars")
  >>> # match all docs where "title" contains the word "Wars"
  >>> F["title"].like("% Wars %")

* Boolean compositions

  * ``&``, ``|``, ``~``: AND, OR and NOT expressions

  >>> # match all records with "rating" >= 4.5 AND "title" starts with "Star Wars"
  >>> e1 = (F["rating"] >= 4.5) & F["title"].startswith("Star Wars")
  >>> # match all records with "director" == "George Lucas" OR "title" starts with "Star Wars"
  >>> e2 = (F["director"] == 'George Lucas') | F["title"].startswith("Star Wars")
  >>> # match all records that are not included in expressions e1 or e2
  >>> e1e2_complement = ~(e1 | e2)

* Field aggregations

  * ``avg``, ``collect``, ``count``, ``countdistinct``, ``max``, ``min``, ``sum``

  >>> # count(*)
  >>> F.count()
  >>> # min(login_time)
  >>> F["login_time"].min()
  >>> # max(login_time) as last_login_time
  >>> F["login_time"].max().named('last_login_time')

* Nested documents and arrays

  * ``[]``: The ``[]`` notation can be used to refer to fields within nested documents and arrays.

  * Consider a collection where documents looked like this example below.

  ::

    {
      "_id": {"u42"},
      "name": {
          "first": "James",
          "middle": "Nicholas",
          "last": "Gray" },
      "tags": [
          "ACID",
          "database locking",
          "two phase commits",
          "five-minute rule",
          "data cube",
          "turing award" ]
     }

  * Example field references to access nested documents and arrays:

  >>> # expression to find all documents where field "name" contains a
  >>> # nested field "middle" with value equal to "Nicholas"
  >>> F["name"]["middle"] == "Nicholas"
  >>>
  >>> # similarly, for array fields, you can specify the array offset.
  >>> # expression to find all documents where the first "tags" field
  >>> # is equal to "ACID"
  >>> F["tags"][0] == "ACID"

  * In order to match against any element within an array field, you can use
    Python's empty slice ``[:]`` notation.

  >>> # expression to find all documents where the "tags" array field
  >>> # contains "ACID" as one of the elements
  >>> F["tags"][:] == "ACID"
  >>> # find all documents where one of the "tags" is "turing award"
  >>> F["tags"][:] == "turing award"


--------------

Query Operator: Filters
-----------------------

=========================
Where operator
=========================
    Syntax:
    ``<Query>.where(<Query>)``

    ``where`` allows you to chain a new query object as a conjuntion.
    In most cases, field reference expressions are sufficient, but
    ``where`` comes in especially handy when you want to sub-select
    documents following another operation such as a sort or an aggregation.

    Examples:
    ::

      # find all "Jim"s who are in the top 100 highest scorers
      Q('players')  \\
      .highest(100, F["score"])  \\
      .where(F["first_name"] == "Jim")

.. automethod:: Query.where

Query Operator: Projection
--------------------------

=========================
Select operator
=========================
    Syntax:
    ``<Query>.select(<field_ref> [, <field_ref> [, ...]])``

    Allows you to specify the fields that you wish to include in the
    query results.

    Examples:
    ::

      Q('authors') \\
      .where(F["last_name"] == "Gray")  \\
      .select(F["first_name"], F["last_name"], F["age"])

.. automethod:: Query.select

Query Operator: Pagination
--------------------------

=========================
Limit operator
=========================
    Syntax:
    ``<Query>.limit(<max_results> [, <skip_count>])``

    Limit operator allows you to perform pagination queries and positional
    filters.

    Examples:
    ::

      # find the "_id" field of the 5 most recently uploaded documents
      # since the default sorting is more recently updated first,
      # this query will simply be:
      Q('uploads').limit(5)

      # fetch a third batch of 100 results, for all users older than 18
      # i.e., skip the first 200 results
      Q('uploads').where(F["age"] >= 18).limit(100, skip=200)

.. automethod:: Query.limit

Query Operator: Sorting
-----------------------

=========================
Highest, Lowest operators
=========================
    Syntax:
    ``<Query>.highest(N, <field_ref> [, <field_ref> [, ...]])``,
    ``<Query>.lowest(N, <field_ref> [, <field_ref> [, ...]])``

    Examples:
    ::

      Q(F["last_name"] == "Gray").highest(5, F["score"], F["first_name"])
      Q(F["last_name"] == "Gray").lowest(10, F["salary"])

.. automethod:: Query.highest
.. automethod:: Query.lowest


Query Operator: Aggregation
---------------------------

====================================================
Aggregate operator and field ref aggregate functions
====================================================
    Syntax:
    ``<Query>.aggregate(<field_ref> [, <field_ref> [, ...]])``

    Field reference objects can also include any of the following aggregation
    functions:

    * ``min``
    * ``max``
    * ``avg``
    * ``sum``
    * ``count``
    * ``countdistinct``
    * ``approximatecountdistinct``
    * ``collect``

    You can also optionally provide a field name alias in the
    field reference using the ``named`` function. This comes in
    especially handy for the aggregated fields.

    Examples:
    ::

      # find min and max salaries broken down by age
      Q('employees').aggregate(F["age"], F["salary"].min(), F["salary"].max())
      # will return documents such as:
      # {"age", "18", "min(salary)": 50000, "max(salary)": 150000}
      # {"age", "19", "min(salary)": 50000, "max(salary)": 152000}

      # example using field name alias
      Q('employees').aggregate(F["age"], F["salary"].avg().named("avg_salary"))
      # will return documents such as:
      # {"age", "18", "avg_salary": 82732}

.. automethod:: Query.aggregate

.. automethod:: FieldRef.min
.. automethod:: FieldRef.max
.. automethod:: FieldRef.avg
.. automethod:: FieldRef.sum
.. automethod:: FieldRef.count
.. automethod:: FieldRef.countdistinct
.. automethod:: FieldRef.approximatecountdistinct
.. automethod:: FieldRef.collect

Query Operator: Joins
---------------------------

=========================
Lookup operator
=========================

    Syntax:
    ``<Query>.lookup(<local_field_ref> \
[, <target_field_ref>] \
[, <target_query_ref>] \
[, <new_field_ref>])``

    Lookup operator allows you to perform a LEFT OUTER JOIN between results
    of the current ``<Query>`` object and the results of the
    ``<target_query_ref>``.

    The LEFT OUTER JOIN operation will be performed on the
    ``<local_field_ref>`` from the results of the current ``<Query>`` object
    and the ``<target_field_ref>`` in ``<target_query_ref>``.

    The output of the JOIN operation will be presented as an array value
    within the ``<new_field_ref>`` in the post JOIN results.

    Examples:
    ::

      # Assume you have following 2 collections:
      #
      # "employees" collection:
      #   {"_collection": "emp", "_id": "e42",
      #       "name": "Jim Gray", "deptId": "d7" }
      #   {"_collection": "emp", "_id": "e43",
      #       "name": "Peter Parker", "deptId": "d8" }
      #   {"_collection": "emp", "_id": "e44",
      #       "name": "Jane Doe", "deptId": null }
      #   {"_collection": "emp", "_id": "e45",
      #       "name": "John Smith" }
      #
      # "dept" collection (only has "d7" but no "d8"):
      #   {"_collection": "dept", "_id": "d7", "name": "eng" }
      #

      # fetch all employees along with their relevant dept records
      rs = Client()
      q = Q("emp").lookup(F["deptId"], target_field=F["_id"],
              target_query=Q("dept"))
      everyone = rs.query(q)

      # the above query will return the following results:
      #
      # [ {"_collection": "emp", "_id": "e42",
      #       "name": "Jim Gray", "deptId": "d7",
      #       "deptId:lookup":
      #           [ {"_collection": "dept", "_id": "d7", "name": "eng"} ]
      #   },
      #   {"_collection": "emp", "_id": "e43",
      #       "name": "Peter Parker", "deptId": "d8",
      #       "deptId:lookup": [ ]
      #   },
      #   {"_collection": "emp", "_id": "e44",
      #       "name": "Jane Doe", "deptId": null,
      #       "deptId:lookup": null
      #   },
      #   {"_collection": "emp", "_id": "e45",
      #       "name": "John Smith"
      #   }
      # ]


    Another example:
    ::

      # another query on the same two collections described above
      # example to fetch name of employee "u42" along with their dept name
      #
      # first define target_query to only select dept id (JOIN field) and name
      target_query = Q("dept").select(F["_id"], F["name"])

      #
      # LEFT OUTER JOIN between employee "u42" and target_query
      # Note: default target_field '_id' works here and is not specified
      jim = Q("emp").where(F["_id"] == "u42").lookup(
                local_field=F["deptId"],
                target_query=target_query,
                new_field=F["dept"])

      #
      # only select relevant fields post JOIN
      jim = jim.select(F["name"], F["dept"])

      # execute the query
      rs = Client()
      everyone = rs.query(jim)

      # the above query will return the following results:
      # [ { "name": "Jim Gray", "dept": [ {"_id": "d7", "name": "eng"} ]
      #   }
      # ]

.. automethod:: Query.lookup

.. _Graph queries:

=========================
Apply operator
=========================
    Syntax:
    ``<Query>.apply(<field_ref>)``

    Apply operator matches all documents whose ``<field_ref>``
    value matches the results of the given ``<Query>``.

    Apply operators enable graph queries that allows you to perform
    graph traversals across related fields in different collections
    without requiring any special data shaping or schema modeling or
    graph specific indexing.

    The common pattern for using ``apply`` will look as follows:

    Step 1: Search, process and filter for relevant documents

    Step 2: Select the field that defines the source vertex

    Step 3: Apply over the field that defines the destination vertex

    Step 4: Go back to Step 1 for another graph traversal hop, if required.

    Examples:
    ::

      # Assume you have following 2 collections:
      #
      # "bad_ips" collection:
      #   {"_collection": "bad_ips", "ip_address": "106.6.6.6", "last_seen": ... }
      #   {"_collection": "bad_ips", "ip_address": "107.6.6.6", "last_seen": ... }
      #
      # "login_attempts" collection:
      #   {"_collection": "login_attempts", "login_ip": "72.43.99.108", ... }
      #   {"_collection": "login_attempts", "login_ip": "106.6.6.6", ... }
      #
      # You can model the above 2 collections as a graph, where there exists
      # an edge between every login_attempts.login_ip and the corresponding
      # bad_ips.ip_address
      #
      # The following query will find all login_attempts from any of the
      # bad_ips:
      rs = Client()
      results = rs.query( Q("bad_ips")  \\
                               .select(F["ip_address"])  \\
                               .apply(F["login_ip"], Q("login_attempts")) )

      # the slow and inefficient way to achive the same would be to
      # materialize the list of all bad_ips and then building out a
      # new query object which could potentially be quite big.
      # DO NOT USE THIS. USE APPLY OPERATOR INSTEAD.
      bad_ips = rs.query(Q("bad_ips"))
      suspicious_logins = None
      for bad_ip in bad_ips:
          is_login_ip_bad = (F["login_ip"] == bad_ip["ip_address"])
          if suspicious_logins:
              suspicious_logins |= is_login_ip_bad
          else:
              suspicious_logins = is_login_ip_bad

      # execute the query against "login_attempts"
      results = rs.query(Q("login_attempts").where(suspicious_logins))


    Another example:
    ::

      # Assume you have following 3 collections:
      # "users"
      # {"_id": "u42", "name": "Patrick", "zip": "94107"}
      # {"_id": "u67", "name": "Paul", "zip": "94306"}
      #
      # "merchants"
      # {"_id": "m2345", "name": "Muddy Waters", "zip": "53706"}
      # {"_id": "m8442", "name": "Beli Deli", "zip": "94002"}
      #
      # "payments"
      # {"_id": "p23rsdkjkapw3", "payer": "u42", "merchant": "m2345", "amount": 17.45, "time": 1488244768}
      # {"_id": "qase8432akdfa", "payer": "u67", "merchant": "m8442", "amount": 42.72, "time": 1488244807}
      # {"_id": "r23raskdfa235", "payer": "u42", "merchant": "m8442", "amount": 11.83, "time": 1488244942}
      # {"_id": "s06slkgeka92s", "payer": "u42", "merchant": "m8442", "amount": 10000, "time": 1488244944}
      #

      # find all merchants in zip code 53706,
      # then find all of their payments,
      # then find the top 100 users by total amount spent
      merchants = Q("merchants").where(F["zip"] == "53706")
      payments  = merchants.select(F["_id"]).apply(F["merchant"], Q("payments"))
      all_users = payments.aggregate(F["payer"], F["amount"].sum().named("total_amount"))
      top_users = all_users.highest(10, F["total_amount"])

      # execute the top_users query
      rs = Client()
      results = rs.query(top_users)

.. automethod:: Query.apply

--------------

Field Expression: Value Comparators
-----------------------------------

=========================
Equality operator: ``==``
=========================
    Supported types: ``int``, ``float``, ``bool``, ``str``

    Syntax:
    ``<field_ref> == <value>``

    Examples:
    ::

      F["first_name"] == "Jim"
      F["year"] == 2017
      F["score"] == 5.0
      F["tags"][:] == 'critical'

===============================================
Value comparators: ``<``, ``<=``, ``>=``, ``>``
===============================================
    Supported types: ``int``, ``float``, ``str``

    Syntax:
    ``<field_ref> < <value>``,
    ``<field_ref> <= <value>``,
    ``<field_ref> >= <value>``,
    ``<field_ref> > <value>``

    Examples:
    ::

      F["year"] < 2000
      F["year"] >= 2007
      F["rating"] >= 4.5
      F["title"] >= "Star Wars"

===============================
Prefix operator: ``startswith``
===============================
    Supported types: ``str``

    Syntax:
    ``<field_ref>.startswith(<prefix>)``

    Examples:
    ::

      F["first_name"].startswith("Ben")

===============================
Like operator: ``like``
===============================
    Supported types: ``str``

    Syntax:
    ``<field_ref>.like(<pattern>)``

    Examples:
    ::

      F["address"].like("%State St%")

===============================
Field alias: ``named``
===============================
    Supported types: All field references

    Syntax:
    ``<field_ref>.named(<new-field-name>)``

    Examples:
    ::

      F["full_name"].named("name")
      F["login_time"].max().named("last_login_time")

===============================
Field existence: ``is_defined``
===============================
    Supported types: All

    Syntax:
    ``<field_ref>.is_defined()``

    Field existence tested with ``<field_ref>.is_defined()`` will match all
    documents where the field is defined, even if it has a null value.

================================
Null comparison: ``is_not_null``
================================
    Supported types: All

    Syntax:
    ``<field_ref>.is_not_null()``

    Field expression ``<field_ref>.is_not_null()`` will match all documents
    where the field is defined and has a non-null value.

.. tip:: There is no ``is_null()`` because of the potential confusion of \
calling ``is_null()`` on an undefined field. Use \
``~<field_ref>.is_not_null()`` or ``<field_ref>.is_defined() & \
~<field_ref>.is_not_null()`` depending on your use case.

=========================
Nested operator
=========================
    Syntax:
    ``<field_ref>.nested(<Query>)``

    ``nested`` operator makes it easy to work with nested array of documents.

    Example:
    ::

        # find all books authored by 'Jim Gray'
        F["authors"].nested((F["first_name") == "Jim") \
& (F["last_name"] == "Gray"))

        # find all users who logged in from given IP on June 06, 2006
        F["logins"].nested((F["ipv4"] == "10.6.6.6") \
& (F["login_date"] == "2006-06-06"))


Field Expression: Boolean Compositions
--------------------------------------

Three different boolean operators (``&``, ``|``, and ``~``) are
overloaded to allow easy composition of boolean operators.

.. note:: The boolean operators are **NOT** ``and``, ``or``, and ``not``, \
as those are special and cannot be overridden in Python.

==================================
AND operator: ``&`` (intersection)
==================================
    Syntax:
    ``<Query object> & <Query object>``

    Examples:
    ::

      # find all documents where field tags contains the "turing award"
      # and the age is greater than 40
      (F["tags"][:] == "turing award") & (F["age"] > 40)

==========================
OR operator: ``|`` (union)
==========================
    Syntax:
    ``<Query object> | <Query object>``

    Examples:
    ::

      # find all documents where the first_name is "jim"
      # or last_name is "gray"
      (F["first_name"] == "jim") | (F["last_name"] == "gray")

==============================
NOT operator: ``~`` (negation)
==============================
    Syntax:
    ``~<Query object>``

    Examples:
    ::

      # find all documents whose title does not contain the term "confidential"
      ~F["title"][:] == "confidential"


"""
import copy
import json
from .value import encode, decode, py_type


class Query(object):
    def __init__(self, source=None, alias=None, child=None, children=None):
        self._source = source
        self._alias = alias

        if child:
            self._source = child.get_source()

        if children:
            for c in children:
                if c.get_source() and not self.get_source():
                    self._source = c.get_source()
                if (
                    self.get_source() and c.get_source() and
                    self.get_source() != c.get_source()
                ):
                    raise ValueError(
                        'Cannot combine multiple sub-queries '
                        'bound to different target collections. '
                    )

        # if query is bound to a collection, then init FieldRef
        # so one can do q.F['name']
        if self.get_source():
            self.F = FieldRef(source=(self.get_alias() or self.get_source()))
        else:
            self.F = FieldRef()

        # init params to empty map
        self.P = ParamDict()  # params explicitly set by user

    def __and__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return AndQuery([self, other])

    def __or__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return OrQuery([self, other])

    def __sub__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return DifferenceQuery(self, other)

    def __invert__(self):
        return NotQuery(self)

    # property getters
    def get_source(self):
        return self._source

    def get_alias(self):
        return self._alias

    def select(self, *fields):
        """ Returns a new query object that when executed will only include
        the list of fields provided as input.

        Args:
            fields (list of FieldRef): fields you wish to select

        Returns:
            Query: new query object that includes the desired field selection
        """
        return SelectQuery(fields, self)

    def where(self, query):
        """ Returns a new query object that when executed will only return
        documents that match the current query object AND the query object
        provided as input.

        Args:
            query (Query): the conjunct query object

        Returns:
            Query: new query object that returns documents in
            **self** AND **query**
        """
        return WhereQuery(query, self)

    def highest(self, limit, *fields):
        """ Returns a new query object that when executed will sort the results
        from the current query object by the list of fields provided as input
        in descending order and return top N defined by the limit parameter.

        Args:
            limit (int): top N results you wish to fetch
            fields (list of FieldRef): fields you wish to sort
            descending by

        Returns:
            Query: new query object that returns top N descending
        """
        return SortQuery('desc', fields, self).limit(limit)

    def lowest(self, limit, *fields):
        """ Returns a new query object that when executed will sort the results
        from the current query object by the list of fields provided as input
        in ascending order and return top N defined by the limit parameter.

        Args:
            limit (int): top N results you wish to fetch
            fields (list of FieldRef): fields you wish to sort
            ascending by

        Returns:
            Query: new query object that returns top N ascending
        """
        return SortQuery('asc', fields, self).limit(limit)

    def aggregate(self, *fields):
        """ Returns a new query object that when executed will aggregate
        results from the current query object by the list of fields
        provided as input.

        Field reference objects can include one of the supported aggregate
        functions such as ``max``, ``min``, ``avg``, ``sum``,
        ``count``, ``countdistinct``, ``approximatecountdistinct``, ``collect``
        as follows: ``<field_ref>.max()``, ``<field_ref>.min()``, ... .

        The list of fields provided as input can contain a mix of field
        references that include an aggregate function and field references
        that does not.

        Args:
            fields (list of FieldRef): fields you wish to aggregate by

        Returns:
            Query: new query object that includes the desired field aggregations
        """
        return AggregateQuery(fields, self)

    def lookup(
        self, local_field, target_field=None, target_query=None, new_field=None
    ):
        """ Lookup allows you to perform a LEFT OUTER JOIN between the results
        of the current query object and the results of the target_query
        provided as input. The LEFT OUTER JOIN operation will be performed
        on the local_field from the results of the current query object
        and the target_field field in target_query. All results from the
        target_query whose target_field value matches the local_field value,
        will be presented as an array value within the new_field in the post
        JOIN results.

        **local_field**:
        For every result document in the current query object, the value of
        the local_field is JOINed against the target_field in the
        target_query. All the documents that match from the target_query
        are presented in a new field whose field name defined by the new_field
        parameter.

        This field is mandatory.

        **target_field**:
        The field from the target_query results against which the local_field
        should be JOINed with.

        This field is optional.

        Default value for target_field is ``F["_id"]`` field.

        **target_query**:
        Defined as a Query object, whose results will be used to JOIN against
        the local_field. The results from the target_query should include
        the target_field.

        All fields selected from the target_query will be present within the
        new_field in the post JOIN results.

        This field is optional.

        Default value for target_query is ``Q(<source-collection>)``, which
        will return all documents in the current collection.

        **new_field**:
        New field in every result document will contain  an array value of all
        the matching results from the target_query.

        If the local field is undefined in a result document,
        then the new field will also be undefined in that result document.

        If the local field value is null in a result document,
        then the new field will also be null in that result document.

        If the local field value is defined and not null in a result document,
        then the new field will have an array of all the documents from the
        target query results whose target field matches the local field value.

        The new field will be an empty array if there are no matches.

        This new field name is optional and defaults to local field name
        concatenated with ``:lookup``.

        .. note:: If the target_query does not contain target_field, then there \
        will not be any matches with the local_field value, and thus the \
        new_field will be an empty array for all results.

        Args:
            local_field (FieldRef): local field you wish to perform the
                LEFT OUTER JOIN.
                This is a required parameter.
            target_field (FieldRef): target field in the target query
                against which you wish to perform the LEFT OUTER JOIN.
                This is an optional parameter.
                Default value: ``F['_id']`` field
            target_query (Query): defines a Query object, whose results will
                be used to match against the local_field.
                This is an optional parameter.
                Defaults to ``Q(<source-collection>)``
            new_field (FieldRef): defines a new field where the results of
                LEFT OUTER JOIN will be present.
                This is an optional parameter.
                Defaults to local field name concatenated with ":lookup".

        Returns:
            Query: new query object that includes the desired LEFT OUTER JOIN
        """
        return LookupQuery(
            local_field,
            new_field=new_field,
            target_field=target_field,
            target_query=target_query,
            child=self
        )

    def apply(self, to_field, target_query=None):
        """ Returns a new query object that when executed will match all
        documents where values of the to_field provided as input will match
        any of the results from the current query object.

        The current query object is expected to have a single field in them,
        and is commonly achieved using the **select** operator.

        Read more about the ``apply`` operator in the `Graph queries`_
        section.

        Args:
            to_fields (FieldRef): field you wish to match the results with
            target_query (Query): Optional. Use this for cross-collection
                graph queries, where the apply needs to work on a different
                collection than what the current query object is referring to.

        Example:
            Find all login IP addresses for user 'u42', and then find all
            activity logs from any of those IP addresses::

                Q('logins')
                  .where(F['user'] == 'u42')
                  .select(F['source_ip'])
                  .apply(F['ip_address'], Q('activity_logs'))


        Returns:
            Query: new query object that includes the desired apply operation
        """
        return ApplyQuery(to_field, self, target_query=target_query)

    def limit(self, limit, skip=0):
        """ Returns a new query object that when executed will only return
        a subset of the results. The query when executed will return no more
        than ``limit`` results after skipping the first ``skip`` number of
        results. The limit operator is most commonly used for pagination.

        Args:
            limit (int): maximum number of results to return
            skip (int): the number of results to skip

        Returns:
            Query: new query object that only returns the desired subset
        """
        return LimitQuery(limit=limit, skip=skip, child=self)

    def sample(self, ratio):
        """ Returns a new query object that when executed will only return
        a uniformly sampled subset of the results. When ratio is 0.1, 1 out of
        every 10 results will be returned.

        Args:
            ratio (float): sampling ratio between 0.0 and 1.0
                0.0 (0% sampling) will not return any results
                1.0 (100% sampling) will return all results

        Returns:
            Query: new query object that only returns the desired subset
        """
        return SampleQuery(ratio=ratio, child=self)

    def boost(self, factor):
        """ Returns a new query object that boosts the results from the given
        query object. Commonly, used along with the the Query.search method,
        to rank certain query contidions higher than other within a single
        search query.

        Eg::

            Q('tv-series').search(
                F["title"].proximity("game of thrones").boost(1.5),
                F["article_body"].proxmity("game of thrones").boost(1.0),
                (F["popular"] == "yes").boost(3.0))

        Args:
            factor (float): boost factor that determines the relevance of the
                matching results.

                Default boost factor is 1.0.
                Need to be a postive float > 0.0, but can be < 1.0.
                Higher boost factors will make matching documents more relevant.

        Returns:
            Query: new query object that incorporate the desired boost
        """
        return BoostedQuery(factor=factor, child=self)

    def search(self, *conditions):
        """ Returns a new query object that performs a search query against
        the given set of conditions.

        Documents that match all the conditions will be considered more
        relevant and should appear before documents that only match
        a subset of the given conditions.

        Unlike boolean AND queries that will only return if all the
        conditions are met, search queries perform a Weak-AND, where if no
        documents match all of the given criteria, then the ones that match
        most of the conditions will be returned.

        Each of the individual conditions could be boosted using the Query.boost
        method to control the relevance of each input query condition.

        Eg::

            Q('tv-series').search(
                F["title"].proximity("game of thrones").boost(1.5),
                F["article_body"].proxmity("game of thrones").boost(1.0),
                (F["popular"] == "yes").boost(3.0))

        Args:
            conditions (list of Query): conditions you want to search against

        Returns:
            Query: new search query object with the given set of conditions
        """
        return SearchQuery(conditions=conditions, child=self)

    def score(self, name, code, context=''):
        """ Returns a new query object that performs custom scoring on
        documents. Each document is added (or modified) a ':score' field
        and sorted based on that field, in descending order.

        The first parameter is name of the custom scorer. The only one currently
        supported is a javascript scorer, with a name 'js'.

        For the javascript scorer, the second parameter is a custom javascript
        code that is executed on each document. The code needs to export a
        function called 'score' that takes two parameters: 1) a document to be
        scored, 2) context (see third parameter). Unfortunatley only
        ECMAScript 5.1 is supported at the moment.

        The third parameter is context, which is passed to the javascript
        function as-is. It can be either a string or a dictionary. Dictionary
        will be converted to JSON when passed to javascript function.

        Eg::

            Q('tv-series').search(
                F["title"].proximity("game of thrones"))
                .score("js",
                       ("var score = function(doc, context) "
                       "{ return doc[':score'] * context['boost']; }"),
                       {"boost": 1.2})

        Args:
            name (str): name of the scorer, currently only "js" is supported
            code (str): In javascript scorer, custom javascript code that
                        scores the document
            context (str or dict): In javascript scorer, context that will be
                                   passed to javascript function

        Returns:
            Query: new score query object with the given set of parameters
        """
        return ScoreQuery(name, code, context, child=self)

    def __str__(self):
        try:
            return self.sql()[0]
        except NotImplementedError:
            return repr(self)

    def sqlprepare(self, sqlsel):
        """ Returns an SQLSelect object, which can be used to build
        the SQL version of the query.
        """
        return sqlsel

    def sqlbuild(self, sqlsel):
        """ Returns an SQLSelect object, which can be used to generate
        the SQL text for the query.
        """
        raise NotImplementedError(
            'Class {} does not implement sqlbuild()'.format(type(self))
        )

    def sqlexpression(self, **kwargs):
        """ Returns a text SQL fragment for the underlying query expression.
        """
        raise NotImplementedError(
            'Class {} does not implement sqlexpression()'.format(type(self))
        )

    def sql(self, **kwargs):
        """ Returns a tuple of (SQL, params) for the underlying query
        expression.
        """
        # step 1: prepare and get all sql params
        params = ParamDict()
        params.update(self.P)
        sqlsel = self.sqlprepare(SQLSelect(params=params))
        sqlsel.P.update(self.P)  # merge with user set params

        # step 2: build the SQL expression into sqlsel
        sqlsel = self.sqlbuild(sqlsel)

        # step 3:
        # translate sqlsel into sql text and return sql params from step 1
        return (sqlsel.sqlexpression(**kwargs), sqlsel.P)


def _sqlexp(s, delim=' ', **kwargs):
    s = decode(s)
    if isinstance(s, list) or isinstance(s, tuple):
        return delim.join([_sqlexp(x, **kwargs) for x in s])
    if isinstance(s, set):
        return delim.join(sorted(set([_sqlexp(x, **kwargs) for x in s])))
    if isinstance(s, bool):
        return str(s).lower()
    if isinstance(s, int) or isinstance(s, float):
        return str(s)
    if isinstance(s, str):
        return json.dumps(s)
    if isinstance(s, slice):
        return Symbol('[*]')
    if isinstance(s, (SubQuery, SQLSelectSubQuery)):
        kwargs['level'] = kwargs.get('level', 0) + 1
    return s.sqlexpression(**kwargs)


def _sqlprep(sqlsel, *args, **kwargs):
    for arg in args:
        if not hasattr(arg, 'sqlprepare'):
            continue
        if not callable(arg.sqlprepare):
            continue
        sqlsel = arg.sqlprepare(sqlsel)
    return sqlsel


def _escape_chars(v, chars=set()):
    for c in set(chars):
        v = v.replace(c, r"\{}".format(c))
    return v


class SQLSelect(object):
    def __init__(self, params=None):
        self._select_list = []
        self._from_list = []
        self._join_list = []
        self._where_list = []
        self._groupby_list = []
        self._orderby_list = []
        self._limit = None
        self._skip = None
        self._build_order = 0
        self.P = params or ParamDict()

    def _set_build_order(self, build_order):
        if self._build_order < build_order:
            self._build_order = build_order
        return self

    def _enforce_build_order(self, build_order, coalesce=False):
        """ use to automatically nest and build inner subquery in
        cases such as Q().aggregate().limit().where()
        """
        if self._build_order < build_order:
            return self._set_build_order(build_order)
        if coalesce and self._build_order == build_order:
            return self
        sq = SQLSelectSubQuery(child=self, alias='subq')
        return (SQLSelect(params=self.P)
            .add_from(sq)
            ._set_build_order(build_order))

    def add_select(self, *f):
        # allow select from any order, so default to current
        build_order = self._build_order
        # unless select list is non-empty, then force a nested sub query
        if len(self._select_list) > 0:
            build_order = 0
        this = self._enforce_build_order(build_order, coalesce=True)
        this._select_list += f
        return this

    def add_from(self, *t):
        this = self._enforce_build_order(1)
        this._from_list += t
        return this

    def add_join(self, r, p, join_type='JOIN'):
        this = self._enforce_build_order(1, coalesce=True)
        this._join_list.append((Symbol(join_type), r, p))
        return this

    def add_left_outer_join(self, r, p):
        return self.add_join(r, p, join_type='LEFT OUTER JOIN')
        return this

    def add_where(self, *op):
        this = self._enforce_build_order(2, coalesce=True)
        this._where_list += op
        return this

    def add_groupby(self, *f):
        this = self._enforce_build_order(3)
        this._groupby_list += f
        return this

    def add_orderby(self, mode, *fields):
        this = self._enforce_build_order(4)
        this._orderby_list += [(field, Symbol(mode)) for field in fields]
        return this

    def add_limit(self, limit, skip=0):
        this = self._enforce_build_order(5)
        this._limit = limit
        this._skip = skip
        return this

    def sqlexpression(self, **kwargs):
        linesep = '\n' + (' ' * kwargs.get('level', 0) * 7)
        ret = ''
        select_list = self._select_list or [F]
        ret += 'SELECT {} '.format(
            _sqlexp(select_list, delim=', ', **kwargs)
        )
        if self._from_list:
            ret += linesep
            ret += '  FROM {} '.format(
                _sqlexp(self._from_list, delim=', ', **kwargs)
            )
        if self._join_list:
            ret += linesep
            for j in self._join_list:
                ret += '    {} {} ON {} '.format(
                    _sqlexp(j[0], **kwargs), _sqlexp(j[1], **kwargs),
                    _sqlexp(j[2], **kwargs)
                )
        if self._where_list:
            ret += linesep
            ret += ' WHERE {} '.format(
                _sqlexp(self._where_list, delim=' AND ', **kwargs)
            )
        if self._groupby_list:
            ret += linesep
            ret += ' GROUP BY {} '.format(
                _sqlexp(self._groupby_list, delim=', ', **kwargs)
            )
        if self._orderby_list:
            ret += linesep
            ret += ' ORDER BY {} '.format(
                _sqlexp(self._orderby_list, delim=', ', **kwargs)
            )
        if self._limit:
            ret += linesep
            ret += ' LIMIT {} '.format(_sqlexp(self._limit, **kwargs))
            if self._skip:
                ret += ' OFFSET {} '.format(_sqlexp(self._skip, **kwargs))
        return ret

    def __str__(self):
        return self.sqlexpression()


class SQLSelectText(SQLSelect):
    def __init__(self, sqltext, alias=None):
        self._sqltext = sqltext
        self._alias = alias or 'subq'
        super(SQLSelectText, self).__init__()

    def _enforce_build_order(self, build_order, coalesce=False):
        # irrespective of build_order, turn sqltext into inner sub query
        sq = SQLSelectSubQuery(child=self, alias=self._alias)
        return SQLSelect(params=self.P).add_from(sq)

    def sqlexpression(self, **kwargs):
        return _sqlexp(Symbol(self._sqltext))


class SQLSelectSubQuery(SQLSelect):
    def __init__(self, child, alias=None):
        self._child = child
        self._alias = alias
        super(SQLSelectSubQuery, self).__init__(params=child.P)

    def sqlexpression(self, **kwargs):
        open_paren = '\n' + ' ' * ((kwargs.get('level', 0) * 7) - 1) + '('
        t = [Symbol(open_paren), self._child, Symbol(')')]
        if self._alias:
            t += [Symbol(' as '), self._alias]
        return _sqlexp(t, delim='', **kwargs)


def _QueryString(qs, alias):
    # detect if input string is collection-name or SQL
    if len(qs.strip().split()) == 1:
        # case 2: exactly one token means it is a resource name
        return QueryStringResource(qs.strip(), alias=alias)
    else:
        # case 3: not s-expr, not resource name -> assume SQL
        return QueryStringSQLText(qs, alias=alias)


class QueryStringResource(Query):
    def __init__(self, qs, alias=None):
        self._resource = qs
        super(QueryStringResource, self).__init__(source=qs, alias=alias)

    def sqlbuild(self, sqlsel):
        return sqlsel.add_from(self)

    def sqlexpression(self, **kwargs):
        t = [self._resource]
        if self.get_alias():
            t += [Symbol('as'), self.get_alias()]
        return _sqlexp(t, **kwargs)


class QueryStringSQLText(Query):
    def __init__(self, qs, alias=None):
        self._sqlexpr = qs
        super(QueryStringSQLText, self).__init__(alias=alias)

    def sqlbuild(self, sqlsel):
        return SQLSelectText(self._sqlexpr, alias=self.get_alias())


class SubQuery(Query):
    def __init__(self, child, alias=None):
        if not isinstance(child, Query):
            raise TypeError('invalid query type ' + type(child))
        super(SubQuery, self).__init__(child=child, alias=alias)
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child)

    def sqlbuild(self, sqlsel):
        return sqlsel.add_from(self)

    def sqlexpression(self, **kwargs):
        childsq = self.child.sqlbuild(SQLSelect(params=self.P))
        sq = SQLSelectSubQuery(child=childsq, alias=self.get_alias())
        return sq.sqlexpression(**kwargs)


class MultiTermQuery(Query):
    def __init__(self, children=[], **kwargs):
        super(MultiTermQuery, self).__init__(children=children, **kwargs)
        self.children = children

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, *self.children)


class AndQuery(MultiTermQuery):
    def __and__(self, other):
        if isinstance(other, AndQuery):
            children = self.children + other.children
            if len(children) == 1:
                return children.pop()
            return AndQuery(children)
        if not isinstance(other, Query):
            return NotImplemented
        if len(self.children) == 0:
            return other
        return AndQuery(self.children + [other])

    def sqlexpression(self, **kwargs):
        return _sqlexp(self.children, delim=' AND ', **kwargs)


class OrQuery(MultiTermQuery):
    def __or__(self, other):
        if isinstance(other, OrQuery):
            children = self.children + other.children
            if len(children) == 1:
                return children.pop()
            return OrQuery(children)
        if not isinstance(other, Query):
            return NotImplemented
        if len(self.children) == 0:
            return other
        return OrQuery(self.children + [other])

    def sqlexpression(self, **kwargs):
        return _sqlexp(self.children, delim=' OR ', **kwargs)


class NotQuery(Query):
    def __init__(self, negated):
        super(NotQuery, self).__init__(child=negated)
        self.negated = negated

    def __invert__(self):
        return self.negated

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.negated)

    def sqlexpression(self, **kwargs):
        return _sqlexp(
            [Symbol('NOT'),
             Symbol('('), self.negated,
             Symbol(')')], **kwargs
        )


class DifferenceQuery(Query):
    def __init__(self, base, diminisher):
        super(DifferenceQuery, self).__init__(child=base)
        self.base = base
        self.diminisher = diminisher


def checked_decode(x, ty=None):
    x = decode(x)
    if ty:
        assert isinstance(x, ty)
    return x


class FieldOpQuery(Query):
    def __init__(self, field, v, op, sqlop=None):
        super(FieldOpQuery, self).__init__()
        self.field = _fref(field)
        self.value = v
        if isinstance(v, (BaseRef, Query)):
            self.value_p = v
        else:
            self.value_p = Literal(v)
        self.op = op
        self.sqlop = sqlop
        self.nested_field_expr = self._init_nested_field_op_query()

    def _init_nested_field_op_query(self):
        if not self.field._is_array_field():
            return None
        # turn F['foo'][:]['bar'][:]['baz'] == 10
        # into F['foo'].nested(F['bar'].nested(F['baz'] == 10))
        fparts = self._split_array_path(self.field)
        fop_q = FieldOpQuery(fparts.pop(), self.value_p, self.op, self.sqlop)
        for fpart in reversed(fparts):
            fop_q = fpart.nested(fop_q)
        return fop_q

    def _split_array_path(self, field):
        parts = []
        tp = FieldRef()
        for fp in field._path():
            if isinstance(fp, slice):
                parts.append(tp)
                tp = FieldRef()
                continue
            tp = tp[fp]
        parts.append(tp)
        return parts

    def sqlprepare(self, sqlsel):
        sqlsel = _sqlprep(sqlsel, self.field, self.value)
        """
        FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME

        Disabling auto-params feature for now.
        Remove this once issue #1506 is fixed.

        FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME FIXME

        if not isinstance(self.value, (BaseRef, Query)):
            # replace value with param ref, stash the value in sqlsel
            pname = sqlsel.P.new_param(self.field)
            sqlsel.P[pname] = self.value
            self.value_p = ParamRef(pname)
        """
        return sqlsel

    def sqlexpression(self, **kwargs):
        if self.nested_field_expr:
            return _sqlexp(self.nested_field_expr, **kwargs)
        return _sqlexp([self.field, Symbol(self.sqlop), self.value_p], **kwargs)


class FieldEqQuery(FieldOpQuery):
    def __init__(self, field, v):
        super(FieldEqQuery, self).__init__(field=field, v=v, op='eq', sqlop='=')


def _make_ne(field, v):
    return NotQuery(FieldEqQuery(field, v))


class FieldIntOpQuery(FieldOpQuery):
    def __init__(self, field, v, op, sqlop):
        super().__init__(field, checked_decode(v, int), op, sqlop)


class FieldFloatOpQuery(FieldOpQuery):
    def __init__(self, field, v, op, sqlop):
        super().__init__(field, checked_decode(v, float), op, sqlop)


class FieldStringOpQuery(FieldOpQuery):
    def __init__(self, field, v, op, sqlop):
        super().__init__(field, checked_decode(v, str), op, sqlop)


class FieldBaseRefOpQuery(FieldOpQuery):
    def __init__(self, field, v, op, sqlop):
        super().__init__(field, checked_decode(v, BaseRef), op, sqlop)


def convert_re(s):
    if hasattr(s, 'flags') and hasattr(s, 'pattern'):
        if s.flags != 0:
            raise ValueError('RegexObjects with flags not supported')
        return s.pattern
    return checked_decode(s, str)


class FieldTermMatchesQuery(FieldOpQuery):
    def __init__(self, field, v):
        super().__init__(field, convert_re(v), 'term_matches')

    def sqlexpression(self, **kwargs):
        raise NotImplementedError(
            'term_regex operator in SQL mode is '
            'not yet implemented!'
        )


class FieldTermPrefixQuery(FieldOpQuery):
    def __init__(self, field, v):
        super().__init__(field, checked_decode(v, str), 'term_prefix')

    def sqlexpression(self, **kwargs):
        raise NotImplementedError(
            'term_prefix operator in SQL mode is '
            'not yet implemented!'
        )


class FieldMatchesQuery(FieldOpQuery):
    def __init__(self, field, v):
        super().__init__(field, convert_re(v), 'regex')

    def sqlexpression(self, **kwargs):
        raise NotImplementedError(
            'regex operator in SQL mode is '
            'not yet implemented!'
        )


class FieldLikeQuery(FieldOpQuery):
    def __init__(self, field, v):
        # v could be str or Literal
        v = decode(v)
        super().__init__(field, v, None, 'LIKE')


class FieldPrefixQuery(FieldOpQuery):
    def __init__(self, field, v):
        # v could be str or Literal
        v = decode(v)
        super().__init__(field, v, 'prefix', None)

        # need to escape '%_' since sql query is re-written with LIKE
        if isinstance(v, str):
            v = _escape_chars(v, '%_')
            v = v + '%'
        elif isinstance(v, Literal):
            v = Literal(_escape_chars(v.value, '%_'))
            v = Literal(v.value + '%')
        self.sql_fop = FieldOpQuery(field, v, None, 'LIKE')

    def sqlprepare(self, sqlsel):
        return self.sql_fop.sqlprepare(sqlsel)

    def sqlexpression(self, **kwargs):
        return self.sql_fop.sqlexpression(**kwargs)


class UnaryOpQuery(Query):
    def __init__(self, field, op, pre_sqlop=None, post_sqlop=None):
        super(UnaryOpQuery, self).__init__()
        self.field = _fref(field)
        self.op = op
        self.pre_sqlop = pre_sqlop
        self.post_sqlop = post_sqlop

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.field)

    def sqlexpression(self, **kwargs):
        exp = []
        self.pre_sqlop and exp.append(Symbol(self.pre_sqlop))
        exp.append(self.field)
        self.post_sqlop and exp.append(Symbol(self.post_sqlop))
        return _sqlexp(exp, **kwargs)


class FieldIsDefinedQuery(UnaryOpQuery):
    def __init__(self, field):
        super(FieldIsDefinedQuery, self).__init__(
            field, 'is_defined', post_sqlop='IS NOT NULL'
        )


class FieldIsNullQuery(UnaryOpQuery):
    def __init__(self, field):
        super(FieldIsNullQuery, self).__init__(
            field, 'not-implemented', post_sqlop='IS NULL'
        )


class FieldExistsQuery(UnaryOpQuery):
    def __init__(self, field):
        super(FieldExistsQuery, self).__init__(
            field, 'not-implemented', pre_sqlop='EXISTS'
        )
        self.field = field


class FieldIsNotNullQuery(UnaryOpQuery):
    def __init__(self, field):
        super(FieldIsNotNullQuery, self).__init__(field, 'is_not_null')

    def sqlexpression(self, **kwargs):
        return _sqlexp(
            FieldIsDefinedQuery(self.field) & FieldOpQuery(
                self.field, Symbol('NULL_VALUE'), 'ne', 'IS DISTINCT FROM'
            ), **kwargs
        )


class ApplyQuery(Query):
    def __init__(self, to_field, child, target_query=None):
        if target_query:
            if not child.get_source():
                raise ValueError(
                    'cannot specify target_query for apply '
                    'when inner query is not bound to a collection'
                )
        super(ApplyQuery, self).__init__(child=(target_query or child))
        self.to_field = _fref(to_field)
        self.child = SubQuery(child)
        self.target_query = target_query or Q(child.get_source())
        self.in_q = FieldOpQuery(self.to_field, self.child, 'apply', 'IN')

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.in_q, self.target_query)

    def sqlbuild(self, sqlsel):
        sqlsel.add_from(self.target_query)
        sqlsel.add_where(self.in_q)
        return sqlsel


class NestedQuery(Query):
    def __init__(self, base_field, inner_query):
        super(NestedQuery, self).__init__(child=inner_query)
        self.base_field = _fref(base_field)
        self.inner_query = inner_query
        uq = Symbol('UNNEST({})'.format(self.base_field))
        self.unnest_query = FieldExistsQuery(
            SubQuery(Q(uq, alias='uq').where(inner_query).select(Literal(1)))
        )

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.base_field, self.unnest_query)

    def sqlexpression(self, **kwargs):
        return _sqlexp(self.unnest_query)


class ProximityQuery(Query):
    def __init__(self, field, search_query, analyzer):
        super(ProximityQuery, self).__init__()
        self.search_field = field
        self.search_query = search_query
        self.analyzer = analyzer


class LookupQuery(Query):
    def __init__(
        self, local_field, target_field, target_query, new_field, child
    ):
        super(LookupQuery, self).__init__(child=child)
        #
        # initialize all fields with appropriate defaults when not specified
        self.local_field = _fref(local_field)
        if target_field is None:
            self.target_field = F['_id']
        else:
            self.target_field = _fref(target_field)
        if new_field is None:
            self.new_field = copy.copy(self.local_field)
            self.new_field._concat(':lookup')
        else:
            self.new_field = _fref(new_field)
        if target_query is None:
            self.target_query = AllQuery(child=child).select(self.target_field)
        else:
            self.target_query = target_query
        self.child = child
        #
        # rewrite SQL components with a nested subquery with group-by
        target_q_alias = (
            self.target_query.get_alias() or self.target_query.get_source()
        )
        if not target_q_alias:
            raise ValueError(
                'Cannot use lookup operator with a target query '
                'that is not bound to a particular collection.'
            )
        sub_q_alias = target_q_alias + '_lookup'
        self.sub_q = SubQuery(
            SubQuery(self.target_query).aggregate(
                self.target_field,
                self.target_query.F.collect().named(self.new_field)
            ),
            alias=sub_q_alias
        )
        self.eq_q = FieldEqQuery(
            self.child.F[self.local_field], self.sub_q.F[self.target_field]
        )

    def sqlprepare(self, sqlsel):
        return _sqlprep(
            sqlsel, self.sub_q, self.eq_q, self.new_field, self.child
        )

    def sqlbuild(self, sqlsel):
        sqlsel = self.child.sqlbuild(sqlsel)
        sqlsel = sqlsel.add_left_outer_join(self.sub_q, self.eq_q)
        sqlsel = sqlsel.add_select(self.child.F, self.sub_q.F[self.new_field])
        return sqlsel


class AllQuery(AndQuery):
    def __init__(self, child=None):
        super(AllQuery, self).__init__(child=child)
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child)

    def sqlbuild(self, sqlsel):
        return self.child.sqlbuild(sqlsel)


class NoneQuery(OrQuery):
    def __init__(self):
        super(NoneQuery, self).__init__()


class AggregateQuery(Query):
    def __init__(self, fields, child):
        super(AggregateQuery, self).__init__(child=child)
        self.dimension_fields = []
        self.aggregate_fields = []
        for f in fields:
            fr = _fref(f)
            if isinstance(fr, AggFieldRef):
                self.aggregate_fields += [fr]
            else:
                self.dimension_fields += [fr]
        self.child = child

    def sqlprepare(self, sqlsel):
        all_fields = self.dimension_fields + self.aggregate_fields
        return _sqlprep(sqlsel, self.child, *all_fields)

    def sqlbuild(self, sqlsel):
        all_fields = self.dimension_fields + self.aggregate_fields
        sqlsel = self.child.sqlbuild(sqlsel)
        sqlsel = sqlsel.add_select(*all_fields)
        sqlsel = sqlsel.add_groupby(*self.dimension_fields)
        return sqlsel


class SortQuery(Query):
    def __init__(self, mode, fields, child):
        super(SortQuery, self).__init__(child=child)
        self.mode = mode
        self.fields = [_fref(f) for f in fields]
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child, *self.fields)

    def sqlbuild(self, sqlsel):
        sqlsel = self.child.sqlbuild(sqlsel)
        return sqlsel.add_orderby(self.mode, *self.fields)


class LimitQuery(Query):
    def __init__(self, limit, skip, child):
        super(LimitQuery, self).__init__(child=child)
        self.max_results = limit
        self.skip_results = skip
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child, self.max_results, self.skip_results)

    def sqlbuild(self, sqlsel):
        sqlsel = self.child.sqlbuild(sqlsel)
        return sqlsel.add_limit(limit=self.max_results, skip=self.skip_results)


class SampleQuery(Query):
    def __init__(self, ratio, child):
        super(SampleQuery, self).__init__(child=child)
        if type(ratio) == int:
            ratio = float(ratio)
        if type(ratio) != float or ratio < 0.0 or ratio > 1.0:
            raise ValueError(
                'SampleQuery specified with an invalid ratio. '
                '{} is not between 0.0 and 1.0'.format(ratio)
            )
        self.ratio = ratio
        self.child = child


class BoostedQuery(Query):
    def __init__(self, factor, child):
        super(BoostedQuery, self).__init__(child=child)
        if type(factor) == int:
            factor = float(factor)
        if type(factor) != float or factor <= 0.0:
            raise ValueError(
                'BoostedQuery specified with an invalid boost. '
                '{} is not a positive float'.format(factor)
            )
        self.factor = factor
        self.child = child


class SearchQuery(Query):
    def __init__(self, conditions, child):
        super(SearchQuery, self).__init__(child=child)
        self.conditions = [c for c in conditions]
        self.child = child


class ScoreQuery(Query):
    def __init__(self, name, code, context, child):
        super(ScoreQuery, self).__init__(child=child)
        self.name = name
        self.code = code
        self.context = context
        self.child = child


class SelectQuery(Query):
    def __init__(self, fields, child):
        super(SelectQuery, self).__init__(child=child)
        if not isinstance(child, Query):
            return NotImplemented
        self.fields = [_fref(f) for f in fields]
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child, *self.fields)

    def sqlbuild(self, sqlsel):
        sqlsel = self.child.sqlbuild(sqlsel)
        return sqlsel.add_select(*self.fields)


class WhereQuery(Query):
    def __init__(self, pred, child):
        super(WhereQuery, self).__init__(child=child)
        if not isinstance(child, Query) or not isinstance(pred, Query):
            return NotImplemented
        self.pred = pred
        self.child = child

    def sqlprepare(self, sqlsel):
        return _sqlprep(sqlsel, self.child, self.pred)

    def sqlbuild(self, sqlsel):
        sqlsel = self.child.sqlbuild(sqlsel)
        return sqlsel.add_where(self.pred)


def _fref(f):
    if isinstance(f, BaseRef):
        return f
    if type(f) is str:
        return FieldRef()[f]
    return NotImplemented


class BaseRef(object):
    def sqlexpression(self, **kwargs):
        raise NotImplementedError(
            'Class {} does not implement sqlexpression()'.format(type(self))
        )


class Symbol(BaseRef):
    def __init__(self, v):
        self.value = v

    def __str__(self):
        return str(self.value)

    def sqlexpression(self, **kwargs):
        return self.value


class Literal(BaseRef):
    def __init__(self, v):
        self.value = v

    def sqlexpression(self, **kwargs):
        if isinstance(self.value, str):
            return "'{}'".format(_escape_chars(self.value, "'"))
        return str(self.value)


class FieldPathPartRef(BaseRef):
    def __init__(self, fp, level=0):
        self.fp = fp
        self.level = level

    def sqlexpression(self, **kwargs):
        if self.level == 0:
            ret = [self.fp]
        elif isinstance(self.fp, slice):
            ret = [Symbol('[*]')]
        elif isinstance(self.fp, int):
            ret = [Symbol('[{}]'.format(self.fp))]
        else:
            ret = [Symbol('.'), self.fp]
        return _sqlexp(ret, delim='', **kwargs)


class FieldRef(BaseRef):
    def __init__(self, name=None, parent=None, source=None):
        if isinstance(name, slice):
            if not (
                name.start is None and name.stop is None and name.step is None
            ):
                raise ValueError('Only empty slices (":") are supported')
        elif isinstance(name, str):
            if name == '*':
                name = slice(None)
        elif isinstance(name, int):
            pass
        elif name is None:
            assert parent is None
            pass
        else:
            raise TypeError('Invalid FieldRef type ' + type(name))

        # if parent is None it means that name also needs to be None
        assert parent is not None or name is None

        self._name = name
        self._parent = parent
        self._source = source

    def __getitem__(self, name):
        if isinstance(name, FieldRef):
            this = self
            for fp in name._path():
                this = FieldRef(fp, this)
        else:
            this = FieldRef(name, self)
        return this

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError('Field reference has no attribute {}', name)
        return FieldRef(name, self)

    def __iter__(self):
        for part in self._path():
            if isinstance(part, slice):
                yield '*'
                continue
            yield part

    def _concat(self, suffix):
        self._name += suffix

    def _path(self):
        r = self
        p = []
        while r._parent is not None:
            p.append(r._name)
            r = r._parent
        p.reverse()
        return p

    def _is_simple_field_lookup(self):
        return isinstance(self._name, str) and len(self._path()) == 1

    def _is_root_field(self):
        if self._name is None and self._parent is None:
            return True
        if isinstance(self._name, slice) and self._parent is not None:
            return self._parent._is_root_field()

    def _is_array_field(self):
        return '*' in list(self)

    def _get_source(self):
        if self._parent:
            return self._parent._get_source()
        return self._source

    def min(self):
        """ Returns a new FieldRef that represents a min() aggregation
        of the given field.

        Returns:
           AggFieldRef: FieldRef object that represents the desired ``min``
           aggregation.
        """
        return AggFieldRef("min", self)

    def max(self):
        """ Returns a new FieldRef that represents a max() aggregation
        of the given field.

        Returns:
           AggFieldRef: FieldRef object that represents the desired ``max``
           aggregation.
        """
        return AggFieldRef("max", self)

    def avg(self):
        """ Returns a new FieldRef that represents an avg() aggregation
        of the given field.

        Returns:
           AggFieldRef: FieldRef object that represents the desired ``avg``
           aggregation.
        """
        return AggFieldRef("avg", self)

    def sum(self):
        """ Returns a new FieldRef that represents a sum() aggregation
        of the given field.

        Returns:
           AggFieldRef: FieldRef object that represents the desired ``sum``
           aggregation.
        """
        return AggFieldRef("sum", self)

    def count(self):
        """ Returns a new FieldRef that represents a count() aggregation
        of the given field.

        Returns:
          AggFieldRef: FieldRef object that represents the desired ``count``
          aggregation.
        """
        return AggFieldRef(
            "count", self, allow_root=True, root_field=FieldRef()
        )

    def approximatecountdistinct(self):
        """ Returns a new FieldRef that represents a approximatecountdistinct()
        aggregation of the given field.

        Returns:
          AggFieldRef: FieldRef object that represents the desired
          ``approximatecountdistinct`` aggregation.
        """
        return AggFieldRef("approximatecountdistinct", self)

    def countdistinct(self):
        """ Returns a new FieldRef that represents a countdistinct()
        aggregation of the given field.

        Returns:
          AggFieldRef: FieldRef object that represents the desired
          ``countdistinct`` aggregation.
        """
        return AggFieldRef("countdistinct", self)

    def collect(self):
        """ Returns a new FieldRef that represents a collect()
        aggregation of the given field.

        Returns:
          AggFieldRef: FieldRef object that represents the desired
          ``collect`` aggregation.
        """
        return AggFieldRef(
            "collect", self, allow_root=True, root_field=FieldRef()[:]
        )

    def apply(self, inner_query):
        """ Returns a new query object that matches all documents where
        the given field matches the results of the inner_query

        E.g: say the inner query returns N documents {r1, r2, ... rN},
        and each of those contain a single field 'f', then the following
        two expressions are equivalent to one another, but the **apply()**
        version is faster, more efficient and requires only one round-trip::

          <field_ref>.apply(inner_query)

          (<field_ref> == r1['f']) | (<field_ref> == r2['f']) \
... | (<field_ref> == rN['f'])

        Args:
            inner_query (Query): query object on whose results you wish \
            to perform the apply operation on.

        Returns:
            Query: query object that represents the desired apply operation
        """
        return ApplyQuery(self, inner_query)

    def nested(self, nested_query):
        """ Returns a new query object that matches all documents
        where the given inner query matches on one or more individual
        nested documents present within the field path of the given field.

        Useful to run complex query expressions on fields that contain
        an nested array of documents.

        Example:
          Say you have a collection where every document describes a book,
          and each document has an "authors" field that is a nested array of
          documents describing each author::

            {"title": "Transaction Processing: Concepts and Techniques",
             "authors": [
                 {"first_name": "Jim", "last_name": "Gray"},
                 {"first_name": "Andreas", "last_name": "Reuter"},
             ],
             "publisher": ... }

        If you want to do find all books where 'Jim Gray' was one of the
        authors, you can use the following query::

            F["authors"].nested((F["first_name"] == 'Jim') \
& (F["last_name"] == 'Gray'))

        Note: Constructing the same query as follows is incorrect::

            # CAUTION: This is not same as the query above
            (F["authors"][:]["first_name"] == 'Jim') \
& (F["authors"][:]["last_name"] == 'Gray')

        The incorrect version will return all books which has at least one
        author with first name 'Jim' and at least one author with last name
        'Gray', but it need not be the same author. A book with two authors
        named 'Jim Smith' and 'Alice Gray' will also match, which is not what
        is intended.

        Args:
            nested_query (Query): query expression to run on every nested \
            document present within the given field path

        Returns:
            Query: query object that represents desired nested operations
        """
        return NestedQuery(self, nested_query)

    def proximity(self, search_query, analyzer="default"):
        """ Returns a new query object that when executed will perform a
        proximity query with the given input search query, tokenized with the
        specified analyzer on the given field.

        Eg: Split the search query "jim phone number" on whitespace and search
        across all documents on the "email_body" field, while ranking documents
        where the search terms appear closer together higher::

            F["email_body"].proximity("jim phone number", \
analyzer="default")

        Proxmity queries matches the query string against the text and adds
        a boost if terms are close together, i.e. "jim phone number" query will
        score "Jim's phone number" higher than "Jim bought a new phone".
        Proximity queries also does spell checking with an English dictionary.

        Args:
            search_query (str): Text search string
            analyzer (str): Name of the analyzer to use on the search_query
                before performing the search. Defaults to "default".

        Returns:
            Query: query object that represents the proximity query
        """
        return ProximityQuery(self, search_query, analyzer)

    def sqlexpression(self, **kwargs):
        # for root_fields just use '*'
        # eg: SELECT count(*) FROM employees;
        terms = []
        if self._get_source():
            terms += [self._get_source(), Symbol('.')]
        if self._is_root_field():
            terms += [Symbol('*')]
        else:
            terms += [
                FieldPathPartRef(p, i) for i, p in enumerate(self._path())
            ]
        return _sqlexp(terms, delim='', **kwargs)

    def __str__(self):
        try:
            return self.sqlexpression()
        except NotImplementedError:
            return repr(self)


class AggFieldRef(BaseRef):
    def __init__(self, agg_op, field, allow_root=False, root_field=None):
        self._field = field
        self._aggregate_op = agg_op
        if not allow_root and self._field._is_root_field():
            raise ValueError(
                'Aggregation function {} is not allowed on '
                'the root field reference'.format(agg_op)
            )
        if allow_root and field._is_root_field() and root_field is not None:
            self._field = root_field
            self._alias = self._aggregate_op
        elif self._field._is_root_field():
            self._alias = self._aggregate_op
        elif self._field._is_simple_field_lookup():
            self._alias = '{}_{}'.format(self._aggregate_op, self._field._name)
        else:
            self._alias = None

    def __str__(self):
        try:
            return self.sqlexpression()
        except NotImplementedError:
            return repr(self)

    def _check_valid_alias(self):
        if self._alias is None:
            raise ValueError(
                'Complex field paths require explicit aliases: ' +
                self._field.sqlexpression()
            )

    def named(self, alias):
        self._alias = alias
        return self

    def sqlexpression(self, **kwargs):
        return _sqlexp(
            [
                Symbol(self._aggregate_op),
                Symbol('('), self._field,
                Symbol(')'),
                Symbol(' as '), self._alias
            ],
            delim='',
            **kwargs
        )


def _gen_param_name(field):
    if isinstance(field, FieldRef):
        fpath = list(field)
        fpath.reverse()
    elif isinstance(field, str):
        fpath = [field]
    else:
        raise TypeError('invalid field type "{}"'.format(type(field)))
    fname = None
    for fname in fpath:
        if (isinstance(fname, int) or (fname == '*') or (fname == '')):
            continue
        # found a good fname in fpath to seed new param; bail
        break

    # a is for apple, p is for param
    fname = fname or 'p'
    # replace all non alphanumeric with '_'
    fname = [c.isalnum() and c or '_' for c in fname]
    # strip all consecutive '_' with a single '_'
    fname = [
        c for i, c in enumerate(fname)
        if ((i == 0) or (c != '_') or (fname[i] != fname[i - 1]))
    ]
    fname = ''.join(fname)
    return fname


class ParamRef(BaseRef):
    def __init__(self, name=''):
        self._name = name
        self._pname = _gen_param_name(name)

    def _symbol(self):
        return ':' + self._pname

    def __getitem__(self, name):
        if self._name != '':
            raise ValueError('nested parameters are not supported')
        return ParamRef(name)

    def __setitem__(self, name, value):
        raise ValueError(
            'cannot assign values to ParamRef. '
            'use ParamDict instead'
        )

    def __str__(self):
        return self._symbol()

    def sqlexpression(self, **kwargs):
        return _sqlexp(Symbol(self._symbol()), **kwargs)


class ParamDict(dict):
    def __setitem__(self, name, value):
        k = _gen_param_name(name)
        v = value
        _ = self._type_map(value)  # ensure value type is supported
        super(ParamDict, self).__setitem__(k, v)

    def _type_map(self, v):
        # map Python types to Rockset type
        if isinstance(v, bool):
            return 'bool'
        elif isinstance(v, int):
            return 'int'
        elif isinstance(v, float):
            return 'float'
        elif isinstance(v, str):
            return 'string'
        raise TypeError(
            'parameter value of type {} is not supported'.format(type(v))
        )

    def new_param(self, field):
        fname = _gen_param_name(field)
        for i in range(1, 1001):
            candidate = fname
            if i > 1:
                candidate += str(i)
            if candidate not in self:
                return candidate
        raise ValueError(
            'too many parameters bound with name "{}"'.format(fname)
        )

    def sqlparams(self):
        params = []
        for k in sorted(self):
            params.append(
                {
                    'name': k,
                    'value': str(self[k]),
                    'type': self._type_map(self[k])
                }
            )
        return params


_arg_ops = [
    ('__eq__', {}, FieldEqQuery),
    ('__ne__', {}, _make_ne),
    ('term_matches', {
        str: FieldTermMatchesQuery
    }, None),
    ('term_startswith', {
        str: FieldTermPrefixQuery
    }, None),
    ('matches', {
        str: FieldMatchesQuery
    }, None),
    ('startswith', {
        str: FieldPrefixQuery,
        Literal: FieldPrefixQuery
    }, None),
    ('like', {
        str: FieldLikeQuery,
        Literal: FieldLikeQuery
    }, None),
]

for op in (('lt', '<'), ('le', '<='), ('gt', '>'), ('ge', '>=')):

    def f_int(field, v, op=op[0], sqlop=op[1]):
        return FieldIntOpQuery(field, v, op, sqlop)

    def f_float(field, v, op=op[0], sqlop=op[1]):
        return FieldFloatOpQuery(field, v, op, sqlop)

    def f_string(field, v, op=op[0], sqlop=op[1]):
        return FieldStringOpQuery(field, v, op, sqlop)

    def f_baseref(field, v, op=op[0], sqlop=op[1]):
        return FieldBaseRefOpQuery(field, v, op, sqlop)

    type_map = {
        int: f_int,
        float: f_float,
        str: f_string,
        FieldRef: f_baseref,
        ParamRef: f_baseref
    }
    _arg_ops.append(('__{}__'.format(op[0]), type_map, None))

_no_arg_ops = [
    ('is_defined', FieldIsDefinedQuery),
    ('is_not_null', FieldIsNotNullQuery),
]


def _agg_ni(self, x, op=op):
    raise TypeError(
        'Operator {} is not supported for aggregated '
        'field references'.format(op)
    )


for op, type_map, default_ctor in _arg_ops:

    def f(self, x, op=op, type_map=type_map, default_ctor=default_ctor):
        ty = py_type(x)
        ctor = type_map.get(ty, default_ctor)
        if ctor is None:
            raise TypeError(
                'Invalid FieldRef op / type combination: {}, {}'.format(op, ty)
            )
        return ctor(self, x)

    setattr(FieldRef, op, f)
    setattr(AggFieldRef, op, _agg_ni)

for op, ctor in _no_arg_ops:

    def f(self, ctor=ctor):
        return ctor(self)

    setattr(FieldRef, op, f)
    setattr(AggFieldRef, op, _agg_ni)

F = FieldRef()
F.__doc__ = """``F`` is a field reference object that helps in building
query expressions natively using Python language expressions.
``F`` uses Python operator overloading heavily and operations on field
references generate Query_ objects that can be used in conjunction with
``Q`` to build compose complex queries."""

P = ParamRef()
P.__doc__ = """``P`` is a query parameter reference object used to define
query parameters within query expressions."""


def Q(query, alias=None):
    """All query objects are constructed using the ``Q(<collection-name>)``
    query builder construct and are then followed by a chain of query
    operators to build the full query expression."""
    if isinstance(query, Query):
        return SubQuery(query, alias=alias)
    elif type(query) == str:
        return _QueryString(query, alias=alias)
    elif isinstance(query, Symbol):
        return QueryStringResource(query, alias=alias)
    return NotImplemented


Q.all = AllQuery()
Q.none = NoneQuery()



"""Integration between SQLAlchemy and Rockset.

Some code based on:
https://github.com/zzzeek/sqlalchemy/blob/rel_0_5/lib/sqlalchemy/databases/sqlite.py
which is released under the MIT license.

https://github.com/dropbox/PyHive
released under Apache License, Version 2.0

TODO
- fix and test get_schema_names
- fix and test _get_table_columns

"""

from __future__ import absolute_import
from __future__ import unicode_literals

from sqlalchemy import exc
from sqlalchemy import types
from sqlalchemy import util
from sqlalchemy.engine import default
from sqlalchemy.sql import compiler
from rockset.sql import dbapi

class _EverythingSet(object):
    """set contains *everything*"""
    def __contains__(self, _):
        return True

class RocksetIdentifierPreparer(compiler.IdentifierPreparer):
    # Just quote everything to make things simpler / easier to upgrade
    reserved_words = _EverythingSet()

_type_map = {
    # TODO : fix Python inferred types
    'bool': types.Boolean,
    'float': types.Float,
    'int': types.Integer,
    'str': types.String,
    'dict': types.JSON,
    'list': types.ARRAY,

    # types returned by the backend
    'bigint': types.BigInteger,
    'varchar': types.String,
    'document': types.JSON,
    'array': types.ARRAY,
}


class RocksetCompiler(compiler.SQLCompiler):
    def visit_char_length_func(self, fn, **kw):
        return 'length{}'.format(self.function_argspec(fn, **kw))


class RocksetTypeCompiler(compiler.GenericTypeCompiler):
    def visit_TEXT(self, type_, **kw):
        if type_.length:
            return 'VARCHAR({:d})'.format(type_.length)
        else:
            return 'VARCHAR'


class RocksetDialect(default.DefaultDialect):
    name = 'rockset'
    driver = 'rest'

    preparer = RocksetIdentifierPreparer
    statement_compiler = compiler.SQLCompiler
    type_compiler = compiler.GenericTypeCompiler

    supports_alter = False
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = True
    description_encoding = None
    supports_native_boolean = True

    @classmethod
    def dbapi(cls):
        return dbapi

    def create_connect_args(self, url):
        kwargs = {
            'api_server': '{}:{}'.format(url.host, url.port or 443),
            'api_key': url.username,
        }
        return ([], kwargs)

    def get_schema_names(self, connection, **kw):
        return [row['Schema'] for row in connection.execute('SHOW SCHEMAS')]

    def _get_table_columns(self, connection, table_name, schema):
        full_table = self.identifier_preparer.quote_identifier(table_name)
        if schema:
            full_table = (self.identifier_preparer.quote_identifier(schema)
                + '.' + full_table)
        try:
            return connection.execute('SHOW COLUMNS FROM {}'.format(full_table))
        except (dbapi.DatabaseError, exc.DatabaseError) as e:
            # Catch exceptions when fetching cursor's description
            # Check if table exists
            msg = hasattr(e, 'message') and e.message or None
            code = hasattr(e, 'code') and e.code or 0
            if 'does not exist' in msg or code == 404:
                raise exc.NoSuchTableError(table_name)
            else:
                raise

    def has_table(self, connection, table_name, schema=None):
        try:
            self._get_table_columns(connection, table_name, schema)
            return True
        except exc.NoSuchTableError:
            return False

    def get_columns(self, connection, table_name, schema=None, **kw):
        rows = self._get_table_columns(connection, table_name, schema)
        result = []
        for row in rows:
            try:
                coltype = _type_map[row['Type']]
            except KeyError:
                util.warn("Did not recognize type '%s' of column '%s'" % (
                    row['Type'], row['Column']))
                coltype = types.NullType
            result.append({
                'name': row['Column'],
                'type': coltype,
                # newer Rockset no longer includes this column
                'nullable': getattr(row, 'Null', True),
                'default': None,
            })
        return result

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        # no support for foreign keys.
        return []

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        # no support for primary keys.
        return {'constrained_columns': ['_id'], 'name': None}

    def get_indexes(self, connection, table_name, schema=None, **kw):
        # no need for primary keys.
        return []

    def get_table_names(self, connection, schema=None, **kw):
        query = 'SHOW TABLES'
        if schema:
            query += (' FROM '
                + self.identifier_preparer.quote_identifier(schema))
        return [row['Table'] for row in connection.execute(query)]

    def do_rollback(self, dbapi_connection):
        # No transactions in Rockset
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        # requests gives back Unicode strings
        return True

    def _check_unicode_description(self, connection):
        # requests gives back Unicode strings
        return True



from rockset.client import *
from rockset.collection import *
from rockset.cursor import *
from rockset.exception import *
from rockset.field_mapping import *
from rockset.query import *
from rockset.source import *
from rockset.value import *

import json
import pkg_resources


def version():
    version_file = pkg_resources.resource_filename('rockset', 'version.json')
    try:
        with open(version_file, 'r') as vf:
            version = json.load(vf).get('version', None)
            if version is None:
                version = pkg_resources.require('rockset')[0].version
    except OSError as e:
        raise FileNotFoundError(
            'could not locate version.json in install dir. '
            'please uninstall and reinstall package "rockset"'
        ) from e

    return version


__all__ = [
    "Client",
    "Collection",
    "Cursor",
    "Exception",
    "FieldMapping"
    "F",
    "FieldRef",
    "P",
    "ParamRef",
    "Q",
    "Query",
    "Source",
]



class TypeUtil:
    TYPE_COLLECTION="COLLECTION"
    TYPE_INTEGRATION="INTEGRATION"

    """
    returns resource type from upstream resource
    """
    @staticmethod
    def parse_resource_type(type):
        if type in ['col', 'collection', 'collections']:
            return TypeUtil.TYPE_COLLECTION
        elif type in ['int', 'integration', 'integrations']:
            return TypeUtil.TYPE_INTEGRATION
        else:
            return None


""" Base class for Collection objects
"""

import time
import datetime

from rockset.cursor import Cursor
from rockset.exception import InputError
from rockset.query import Query

from rockset.swagger_client.api import CollectionsApi

class Resource(object):
    # instance methods
    def __init__(self, client, name, **kwargs):
        """Represents a single Rockset collection"""
        self.client = client
        self.workspace = 'commons'
        self.name = name
        self.dropped = False
        for key in kwargs:
            setattr(self, key, kwargs[key])
        return
    def __str__(self):
        """Converts the collection into a user friendly printable string"""
        return str(vars(self))
    def asdict(self):
        d = vars(self)
        d.pop('client')
        if not self.dropped:
            d.pop('dropped')
        return d
    def describe(self):
        kwargs = {}
        return CollectionsApi(self.client).get(collection=self.name)

    def drop(self):
        self.dropped = True
        CollectionsApi(self.client).delete(collection=self.name)
        return

    def query(self, q, **kwargs):
        return self.client.query(q=q, collection=self.name, **kwargs)

__all__ = [
    'Resource',
]



from .command_rest import RESTCommand
from .util.type_util import TypeUtil

from docopt import docopt

class Describe(RESTCommand):
    def usage(self):
        return """
usage: rock describe [-ah] <resource-type> <name> ...

Show details about collections or integrations.

arguments:
    <name>                name of the collection or integration
    <resource-type>       oneof collections or integrations.

Valid resource types:
  * collections (aka 'col')
  * integrations (aka 'int')

options:
    -a, --all           display extended stats
    -h, --help          show this help message and exit"""

    def parse_args(self, args):
        parsed_args = dict(docopt(self.usage(), argv=args))
        if parsed_args['--help']:
            ret = self.usage()
            raise SystemExit(ret.strip())

        resource_type = TypeUtil.parse_resource_type(parsed_args['<resource-type>'])
        if resource_type is None:
            ret = 'Error: invalid resource type "{}"\n'.format(resource_type)
            ret += self.usage()
            raise SystemExit(ret.strip())
        return {"resource": {'type': resource_type, 'name': parsed_args['<name>']}}

    def go(self):
        self.logger.info('describe {}'.format(self.resource))
        if self.resource["type"] == TypeUtil.TYPE_COLLECTION:
            return self.go_collection()
        elif self.resource["type"] == TypeUtil.TYPE_INTEGRATION:
            return self.go_integration()
        else:
            return 1

    def go_collection(self):
        for name in self.resource['name']:
            path = '/v1/orgs/{}/ws/{}/collections/{}'.format(
                'self', 'commons', name
            )
            deets = self.get(path)
            if 'sources' in deets:
                nsrcs = []
                for src in deets['sources']:
                    nsrcs.append({k: v for k, v in src.items() if v})
                deets['data']['sources'] = nsrcs
            desc = {}
            if 'data' in deets:
                desc = {k: v for k, v in deets['data'].items() if v}
            self.print_list_yaml(0, [desc])
        return 0

    def go_integration(self):
        for name in self.resource['name']:
            path = '/v1/orgs/{}/integrations/{}'.format(
                'self', name
            )
            response = self.get(path)
            if 'resources' in response:
                nsrcs = []
                for src in response['resources']:
                    nsrcs.append({k: v for k, v in src.items() if v})
                response['data']['resources'] = nsrcs
            desc = {}
            if 'data' in response:
                desc = {k: v for k, v in response['data'].items() if v}
            self.print_list_yaml(0, [desc])
        return 0



from docopt import docopt

from .command_rest import RESTCommand


class List(RESTCommand):
    @classmethod
    def what(cls):
        return ('ls', 'List all collections')

    def usage(self):
        return """
usage: rock ls [-h] [<resource-type>]

List all collections or integrations.

arguments:
    <resource-type>       oneof collections or integrations. Default: collections

Valid resource types:
  * collections (aka 'col')
  * integrations (aka 'int')

options:
  -h, --help            show this help message and exit
        """

    def parse_args(self, args):
        parsed_args = dict(docopt(self.usage(), argv=args, help=False))

        # handle help
        if parsed_args['--help']:
            ret = self.usage()
            raise SystemExit(ret.strip())

        resource_type = parsed_args['<resource-type>']

        # check if list integrations
        if resource_type in ['col', 'collection', 'collections']:
            return {'resource': 'COLLECTIONS'}
        elif resource_type in ['int', 'integration', 'integrations']:
            return {'resource': 'INTEGRATIONS'}
        elif resource_type is not None:
            ret = 'Error: invalid resource type "{}"\n'.format(resource_type)
            ret += self.usage()
            raise SystemExit(ret.strip())

        return {'resource': 'COLLECTIONS'}

    def go(self):
        if self.resource == 'INTEGRATIONS':
            return self.go_integrations()
        return self.go_collections()

    def go_collections(self):
        path = '/v1/orgs/{}/ws/{}/collections'.format('self', 'commons')
        items = self.get(path)['data']
        for item in items:
            item['type'] = 'COLLECTION'
        sorted_items = sorted(items, key=lambda k: k['type'] + ':' + k['name'])
        return (0, sorted_items)

    def go_integrations(self):
        path = '/v1/orgs/{}/integrations'.format('self')
        items = self.get(path)['data']
        for item in items:
            item['type'] = 'AWS'
        sorted_items = sorted(items, key=lambda k: k['type'] + ':' + k['name'])
        return (0, sorted_items)

    def print_result(self, result):
        if self.resource == 'INTEGRATIONS':
            self.print_list(
                0, result, ['type', 'name', 'description', 'created_by']
            )
            return
        self.print_list(
            0, result, ['name', 'status', 'description', 'created_by']
        )
